import re
import io
import time
import os.path
import typing
import tarfile
import asyncio
import logging
import tempfile
import functools
import configparser

import yaml

from ..abc import Service
from ..config import Config
from ..log import LOG_NOTICE
from .item import LibraryItem
from ..application import Application
from .providers.abc import LibraryProviderABC

#

L = logging.getLogger(__name__)


#


class LibraryService(Service):
    """
    Configuration:

    ```ini
    [library]
    providers:
            provider+1://
            provider+2://
            provider+3://
    ```

    The order of providers is important, the priority (or layering) is top-down.

    Each library provider is specified by URL/URI schema:

    * `zk://` or `zookeeper://` for ZooKeeper provider
    * `file://` or local path for FileSystem provider
    * `azure+https://` for Microsoft Azure Storage provider.
    * `git+https://` for Git provider.
    * `libsreg+https://` for Libraries provider.

    The first provider is responsible for providing `/.disabled.yaml`.

    A library is created in “not ready” state, each provider then informs the library when it is ready
    (eg. Zookeeper provider needs to connect to Zookeeper servers). Only after all providers are ready, the library itself becomes ready.
    The library indicates that by the PubSub event `Library.ready!`.
    """

    def __init__(
        self,
        app: Application,
        service_name: str,
        paths: typing.Union[str, typing.List[str], None] = None,
    ):
        """
        Initialize the LibraryService.

        The library service is designed to "exist" in multiple instances,
        with different `paths` setups.
        For that reason, you have to provide unique `service_name`
        and there is no _default_ value for that.

        If `paths` are not provided, they are fetched from `[library]providers` configuration.

        Args:
                app: The ASAB Application.
                service_name: A unique name of the service.
                paths (str | list[str] | None ): Either single path or list of paths with which LibraryService is connected.
        """

        super().__init__(app, service_name)
        self.Libraries: list[LibraryProviderABC] = []
        self.Disabled: dict = {}
        self.DisabledPaths: list = []

        if paths is None:
            # load them from configuration
            try:
                paths = Config.getmultiline("library", "providers")
            except configparser.NoOptionError:
                L.critical(
                    "'providers' option is not present in configuration section 'library'."
                )
                raise SystemExit("Exit due to a critical configuration error.")

        # paths can be string if specified as argument
        if isinstance(paths, str):
            paths = re.split(r"\s+", paths)

        for layer, path in enumerate(paths):
            # Create library for each layer of paths
            self._create_library(path, layer)

        app.PubSub.subscribe("Application.tick/60!", self._on_tick60)

    async def finalize(self, app):
        while len(self.Libraries) > 0:
            lib = self.Libraries.pop(-1)
            await lib.finalize(self.App)

    async def _on_tick60(self, message_type):
        await self._read_disabled()

    def _create_library(self, path, layer):
        library_provider = None
        if path.startswith("zk://") or path.startswith("zookeeper://"):
            from .providers.zookeeper import ZooKeeperLibraryProvider

            library_provider = ZooKeeperLibraryProvider(self, path, layer)

        elif (
            path.startswith("./") or path.startswith("/") or path.startswith("file://")
        ):
            from .providers.filesystem import FileSystemLibraryProvider

            library_provider = FileSystemLibraryProvider(self, path, layer)

        elif path.startswith("azure+https://"):
            from .providers.azurestorage import AzureStorageLibraryProvider

            library_provider = AzureStorageLibraryProvider(self, path, layer)

        elif path.startswith("git+"):
            from .providers.git import GitLibraryProvider

            library_provider = GitLibraryProvider(self, path, layer)

        elif path.startswith("libsreg+"):
            from .providers.libsreg import LibsRegLibraryProvider

            library_provider = LibsRegLibraryProvider(self, path, layer)

        elif path == "" or path.startswith("#") or path.startswith(";"):
            # This is empty or commented line
            return

        else:
            L.error("Incorrect/unknown provider for '{}'".format(path))
            raise SystemExit("Exit due to a critical configuration error.")

        self.Libraries.append(library_provider)

    def is_ready(self) -> bool:
        """
        Check if all the libraries are ready.

        Returns:
                True if all libraries are ready, otherwise False.
        """
        if len(self.Libraries) == 0:
            return False

        return functools.reduce(
            lambda x, provider: provider.IsReady and x, self.Libraries, True
        )

    async def _set_ready(self, provider):
        if len(self.Libraries) == 0:
            return

        if (provider == self.Libraries[0]) and provider.IsReady:
            await self._read_disabled()

        if self.is_ready():
            L.log(LOG_NOTICE, "is ready.", struct_data={"name": self.Name})
            self.App.PubSub.publish("Library.ready!", self)
        elif not provider.IsReady:
            L.log(LOG_NOTICE, "is NOT ready.", struct_data={"name": self.Name})
            self.App.PubSub.publish("Library.not_ready!", self)

    async def read(
        self, path: str, tenant: typing.Optional[str] = None
    ) -> typing.Optional[typing.IO]:
        """
        Read the content of the library item specified by `path`. This method can be used only after the Library is ready.

        Args:
                path (str): Path to the file, `LibraryItem.name` can be used directly.
                tenant (str | None): The tenant to apply. If not specified, the global access is assumed.

        Returns:
                ( IO | None ): Readable stream with the content of the library item. `None` is returned if the item is not found or if it is disabled (either globally or for the specified tenant).

        Examples:

        ```python
        itemio = await library.read('/path', 'tenant')
        if itemio is not None:
                with itemio:
                        return itemio.read()
        ```
        """
        # item path must start with '/'
        assert (
            path[:1] == "/"
        ), "Item path must start with a forward slash (/). For example: /library/Templates/item.json"
        # Item path must end with the extension
        assert (
            len(os.path.splitext(path)[1]) > 0
        ), "Item path must end with an extension. For example: /library/Templates/item.json"

        if self.check_disabled(path, tenant=tenant):
            return None

        for library in self.Libraries:
            itemio = await library.read(path)
            if itemio is None:
                continue
            return itemio

        return None

    async def list(
        self,
        path: str = "/",
        tenant: typing.Optional[str] = None,
        recursive: bool = False,
    ) -> typing.List[LibraryItem]:
        """
        List the directory of the library specified by the path that are enabled for the specified tenant. This method can be used only after the Library is ready.

        Args:
                path (str): Path to the directory.
                tenant (str | None): If specified, items that are enabled for the tenant are filtered.
                recursive (bool): If `True`, return a list of items located at `path` and its subdirectories.

        Returns:
                List of items that are enabled for the tenant.
        """

        # Directory path must start with '/'
        assert (
            path[:1] == "/"
        ), "Directory path must start with a forward slash (/). For example: /library/Templates/"
        # Directory path must end with '/'
        assert (
            path[-1:] == "/"
        ), "Directory path must end with a forward slash (/). For example: /library/Templates/"
        # Directory path cannot contain '//'
        assert (
            "//" not in path
        ), "Directory path cannot contain double slashes (//). Example format: /library/Templates/"

        # List requested level using all available providers
        items = await self._list(path, tenant, providers=self.Libraries)

        if recursive:
            # If recursive scan is requested, then iterate thru list of items
            # find 'dir' types there and list them.
            # Output of this list is attached to the list for recursive scan
            # and also to the final output
            recitems = list(items[:])

            while len(recitems) > 0:
                item = recitems.pop(0)
                if item.type != "dir":
                    continue

                child_items = await self._list(
                    item.name, tenant, providers=item.providers
                )
                items.extend(child_items)
                recitems.extend(child_items)
        return items

    async def _list(self, path, tenant, providers):
        # Execute the list query in all providers in-parallel
        result = await asyncio.gather(
            *[library.list(path) for library in providers], return_exceptions=True
        )

        items = []
        uniq = dict()
        for ress in result:
            if isinstance(ress, KeyError):
                # The path doesn't exists in the provider
                continue

            if isinstance(ress, Exception):
                raise ress

            for item in ress:
                item.disabled = self.check_disabled(item.name, tenant=tenant)

                # If the item already exists, merge or override it
                pitem = uniq.get(item.name)
                if pitem is not None:
                    pitem = uniq[item.name]
                    if pitem.type == "dir" and item.type == "dir":
                        # Directories are joined
                        pitem.providers.extend(item.providers)
                    elif pitem.type == "item":
                        for i, provider in enumerate(providers):
                            if provider in item.providers:
                                index = i
                                break
                        pitem.override = index
                # Other item types are skipped
                else:
                    uniq[item.name] = item
                    items.append(item)
        items.sort(key=lambda x: x.name)
        return items

    async def _read_disabled(self):
        # `.disabled.yaml` is read from the first configured library
        # It is applied on all libraries in the configuration.
        disabled = await self.Libraries[0].read("/.disabled.yaml")

        if disabled is None:
            self.Disabled = {}
            self.DisabledPaths = []
            return

        try:
            disabled = yaml.safe_load(disabled)
        except Exception:
            self.Disabled = {}
            self.DisabledPaths = []
            L.exception("Failed to parse '/.disabled.yaml'")
            return

        if disabled is None:
            self.Disabled = {}
            self.DisabledPaths = []
            return

        if isinstance(disabled, set):
            # This is for a backward compatibility (Aug 2023)
            self.Disabled = {key: "*" for key in self.Disabled}
            self.DisabledPaths = []
            return

        self.Disabled = {}
        self.DisabledPaths = []
        for k, v in disabled.items():
            if k.endswith("/"):
                self.DisabledPaths.append((k, v))
            else:
                self.Disabled[k] = v

        # Sort self.DisabledPaths from the shortest to longest
        self.DisabledPaths.sort(key=lambda x: len(x[0]))

    def check_disabled(self, path: str, tenant: typing.Optional[str] = None) -> bool:
        """
        Check if the item specified in path is disabled, either globally or for the specified tenant.

        Args:
                path (str): Path to the item to be checked.
                tenant (str | None): The tenant to apply. If not specified, the global access is assumed.

        Returns:
                `True` if the item is disabled for the tenant.
        """
        if not isinstance(path, str) or not path:
            raise ValueError("The 'path' must be a non-empty string.")

        # First check disabled by path
        for dp, disabled in self.DisabledPaths:
            if path.startswith(dp):
                if "*" in disabled:
                    # Path is disabled for everybody
                    return True

                if tenant is not None and tenant in disabled:
                    # Path is disabled for a specified tenant
                    return True

        # Then check for a specific item entries

        disabled = self.Disabled.get(path)

        if disabled is None:
            return False

        if "*" in disabled:
            # Item is disabled for everybody
            return True

        if tenant is not None and tenant in disabled:
            # Item is disabled for a specified tenant
            return True

        return False

    async def export(
        self,
        path: str = "/",
        tenant: typing.Optional[str] = None,
        remove_path: bool = False,
    ) -> typing.IO:
        """
        Return a file-like stream containing a gzipped tar archive of the library contents of the path.

        Args:
                path: The path to export.
                tenant (str | None ): The tenant to use for the operation.
                remove_path: If `True`, the path will be removed from the tar file.

        Returns:
                A file object containing a gzipped tar archive.
        """

        # Directory path must start with '/'
        assert (
            path[:1] == "/"
        ), "Directory path must start with a forward slash (/). For example: /library/Templates/"
        # Directory path must end with '/'
        assert (
            path[-1:] == "/"
        ), "Directory path must end with a forward slash (/). For example: /library/Templates/"
        # Directory path cannot contain '//'
        assert (
            "//" not in path
        ), "Directory path cannot contain double slashes (//). Example format: /library/Templates/"

        fileobj = tempfile.TemporaryFile()
        tarobj = tarfile.open(name=None, mode="w:gz", fileobj=fileobj)

        items = await self._list(path, tenant, providers=self.Libraries[:1])
        recitems = list(items[:])

        while len(recitems) > 0:
            item = recitems.pop(0)
            if item.type != "dir":
                continue

            child_items = await self._list(item.name, tenant, providers=item.providers)
            items.extend(child_items)
            recitems.extend(child_items)

        for item in items:
            if item.type != "item":
                continue
            my_data = await self.Libraries[0].read(item.name)
            if remove_path:
                assert item.name.startswith(path)
                tar_name = item.name[len(path) :]
            else:
                tar_name = item.name
            info = tarfile.TarInfo(tar_name)
            my_data.seek(0, io.SEEK_END)
            info.size = my_data.tell()
            my_data.seek(0, io.SEEK_SET)
            info.mtime = time.time()
            tarobj.addfile(tarinfo=info, fileobj=my_data)

        tarobj.close()
        fileobj.seek(0)
        return fileobj

    async def subscribe(self, paths: typing.Union[str, typing.List[str]]) -> None:
        """
        Subscribe to changes for specified paths of the library.

        In order to notify on changes in the Library, this method must be used after the Library is ready.

        Args:
                paths (str | list[str]): Either single path or list of paths to be subscribed. All the paths must be absolute (start with '/').

        Examples:
        ```python
        class MyApplication(asab.Application):

                async def initialize(self):
                        self.PubSub.subscribe("Library.ready!", self.on_library_ready
                        self.PubSub.subscribe("Library.change!", self.on_library_change)

                async def on_library_ready(self, event_name, library=None):
                        await self.LibraryService.subscribe(["/alpha","/beta"])

                def on_library_change(self, message, provider, path):
                        print("New changes in the library found by provider: '{}'".format(provider))

        ```
        """
        if isinstance(paths, str):
            paths = [paths]
        for path in paths:
            assert (
                path[:1] == "/"
            ), "Absolute path must be used when subscribing to the library changes"

            for provider in self.Libraries:
                await provider.subscribe(path)



import abc


class Module(abc.ABC):
    """
    Abstract class for ASAB modules.

    Modules are registered at the module registry `asab.Application.Modules`, managed by an application object.
    Module can be loaded by ASAB and typically provides one or more Service objects.

    Every module provides asynchronous methods `initialize()` and `finalize()`.
    `initialize()` is called in the very beginning of the run-time,
    `finalize()` in the gentle shut-down of the application.

    Examples:

    Recommended structure of the ASAB module:

    ```
    my_module/
            - __init__.py
            - my_service.py
    ```

    Content of `__init__.py`:

    ```python title="__init__.py"
    import bspump.asab as asab
    from .my_service import MyService

    # Extend ASAB configuration defaults
    asab.Config.add_defaults({
            'my_module': {
                    'foo': 'bar'
            }
    })

    class MyModule(asab.Module):
            def __init__(self, app):
                    super().__init__(app)
                    self.service = MyService(app, "MyService")
    ```

    And this is how the module is loaded:

    ```python
    from mymodule import MyModule
    ...
    app.add_module(MyModule)
    ```

    """

    def __init__(self, app):
        pass

    # Lifecycle

    async def initialize(self, app):
        """
        This method is called when the Module is initialized. It can be overridden by an user.

        Args:
                app (asab.Application): Reference to ASAB application.
        """
        pass

    async def finalize(self, app):
        """
        This method is called when the Module is finalized, e.g., during application `exit-time`. It can be overridden by an user.

        Args:
                app (asab.Application): Reference to ASAB application.
        """
        pass



import logging

from .namedmatrix import NamedMatrix, PersistentNamedMatrix

###

L = logging.getLogger(__name__)

###


class SessionMatrix(NamedMatrix):
    ConfigDefaults = {"primary_name": "id"}

    """
		Matrix, specific for `SessionAnalyzer`.

	"""

    def __init__(self, app, dtype="float_", id=None, config=None):
        if not isinstance(dtype, str):
            dtype = dtype[:]
        super().__init__(app, dtype=dtype, id=id, config=config)
        self.PrimaryName = self.Config["primary_name"]

    def store(self, row_name: str, event):
        """
        Store the event in the matrix.
        The event must prepared so that it matches a data type of the cell (dtype)
        """
        row_index = self.get_row_index(row_name)
        if row_index is None:
            return False
        self.Array[row_index] = event

    def store_event(self, row_index: int, event, keys=None):
        if keys is None:
            keys = event.keys()

        names = self.Array.dtype.names
        if names is None:
            L.warning("The matrix does not have correct column-like dtype")
            raise

        for key in keys:
            if key in names:
                self.Array[row_index][key] = event[key]

    def decode_row(self, row_index: int, keys=None):
        if keys is None:
            keys = list(self.Array.dtype.names)

        if keys is None:
            L.warning("The matrix does not have correct column-like dtype")
            raise

        event = dict(zip(keys, self.Array[row_index][keys]))
        event[self.PrimaryName] = self.get_row_name(row_index)

        return event


class PersistentSessionMatrix(PersistentNamedMatrix):
    ConfigDefaults = {"primary_name": "id"}

    def __init__(self, app, dtype="float_", id=None, config=None):
        if not isinstance(dtype, str):
            dtype = dtype[:]
        super().__init__(app, dtype=dtype, id=id, config=config)
        self.PrimaryName = self.Config["primary_name"]

    def store(self, row_name: str, event):
        """
        Store the event in the matrix.
        The event must prepared so that it matches a data type of the cell (dtype)
        """
        row_index = self.get_row_index(row_name)
        if row_index is None:
            return False
        self.Array[row_index] = event

    def store_event(self, row_index: int, event, keys=None):
        if keys is None:
            keys = event.keys()

        names = self.Array.dtype.names
        if names is None:
            L.warning("The matrix does not have correct column-like dtype")
            raise

        for key in keys:
            if key in names:
                self.Array[row_index][key] = event[key]

    def decode_row(self, row_index: int, keys=None):
        if keys is None:
            keys = list(self.Array.dtype.names)

        if keys is None:
            L.warning("The matrix does not have correct column-like dtype")
            raise

        event = dict(zip(keys, self.Array[row_index][keys]))
        event[self.PrimaryName] = self.get_row_name(row_index)

        return event



import logging

from ..abc.sink import Sink

L = logging.getLogger(__name__)


class MQTTSink(Sink):
    ConfigDefaults = {"topic": "default", "qos": 0, "retain": False}

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Connection = pipeline.locate_connection(app, connection)

    def process(self, context, event):
        self.Connection.publish_to_topic(
            self.Config["topic"], event, self.Config["qos"], self.Config["retain"]
        )



from ...abc import Expression

from ..value.valueexpr import VALUE


class SELF(Expression):
    Attributes = {
        "Event": ["*"],  # Must of the the same type as the whole expression
    }

    Category = "Statements"

    def __init__(self, app, *, arg_event):
        super().__init__(app)

        if isinstance(arg_event, Expression):
            self.Event = arg_event
        else:
            self.Event = VALUE(app, value=arg_event)

        self.Self = None  # Will be filled later

    def initialize(self, self_expression):
        self.Self = self_expression

    def __call__(self, context, event, *args, **kwargs):
        new_event = self.Event(context, event, *args, **kwargs)
        return self.Self(context, new_event, *args, **kwargs)

    def get_outlet_type(self):
        return self.Self.get_outlet_type()



from .abc import Expression, SequenceExpression
from .expression.value.valueexpr import VALUE


def declaration_to_dot(decl, fname):
    with open(fname, "w") as fo:
        return declaration_to_dot_stream(decl, fo)


def declaration_to_dot_stream(decl, ostream):
    nodes = set()
    links = set()

    ostream.write(
        'digraph G {\n\trankdir=LR;\n\tgraph [fontname = "helvetica"];\n\tnode [fontname = "helvetica"];\n\tedge [fontname = "helvetica"];\n'
    )

    for parent, key, obj in decl.walk():
        if parent is not None:
            nodes.add(parent)

        if isinstance(obj, Expression):
            nodes.add(obj)

            if (parent is not None) and (parent.Id, obj.Id, key) not in links:
                if obj in nodes:
                    style = "dashed"  # Link by alias
                else:
                    style = "filled"

                ostream.write(
                    '\t"{}" -> "{}" [style={},label="{}"];\n'.format(
                        parent.Id, obj.Id, style, key
                    )
                )
                links.add((parent.Id, obj.Id, key))

        else:
            vid = ".{}.val".format(parent.Id)
            if (parent.Id, vid, key) not in links:
                ostream.write(
                    '\t"{}" -> "{}" [label="{}"];\n'.format(parent.Id, vid, key)
                )
                ostream.write(
                    '\t"{}" [shape=box,label="<{}>\\n{}"]\n'.format(
                        vid, obj.__class__.__name__, obj
                    )
                )
                links.add((parent.Id, vid, key))

    for obj in nodes:
        if isinstance(obj, SequenceExpression):
            shape = "hexagon"
            addinfo = ""
        elif isinstance(obj, VALUE):
            shape = "box"
            addinfo = "\\n" + str(obj.Value)
        else:
            shape = "octagon"
            addinfo = ""

        ostream.write(
            '\t"{}" [label="<{}>\\n{}{}",shape="{}",style=filled,fillcolor=lightgray];\n'.format(
                obj.Id, obj.get_outlet_type(), obj.__class__.__name__, addinfo, shape
            )
        )

    ostream.write("}\n")



import configparser
import logging
import asyncio
import os

from ..config import Config
from ..abc import Service
from .metrics import (
    Metric,
    Counter,
    EPSCounter,
    Gauge,
    DutyCycle,
    AggregationCounter,
    Histogram,
    CounterWithDynamicTags,
    AggregationCounterWithDynamicTags,
    HistogramWithDynamicTags,
)
from .storage import Storage


#

L = logging.getLogger(__name__)

#


class MetricsService(Service):
    def __init__(self, app, service_name):
        super().__init__(app, service_name)

        self.Metrics = []
        self.Targets = []
        self.Tags = {
            "host": app.HostName,
            "appclass": app.__class__.__name__,
        }
        if app.DeploymentId:
            self.Tags["deployment_id"] = app.DeploymentId

        # A identified of the host machine (node); added if available at environment variables
        node_id = os.getenv("NODE_ID", None)
        if node_id is not None:
            self.Tags["node_id"] = node_id

        service_id = os.getenv("SERVICE_ID", None)
        if service_id is not None:
            self.Tags["service_id"] = service_id

        # A unique identifier of a microservice; added as an environment variable.
        instance_id = os.getenv("INSTANCE_ID", None)
        if instance_id is not None:
            self.Tags["instance_id"] = instance_id

        self.Storage = Storage()

        app.PubSub.subscribe("Application.tick/60!", self._on_flushing_event)

        if Config.has_option("asab:metrics", "target"):
            for target in Config.get("asab:metrics", "target").split():
                target = target.strip()
                try:
                    target_type = Config.get("asab:metrics:{}".format(target), "type")
                except configparser.NoOptionError:
                    # This allows to specify the type of the target by its name
                    target_type = target

                if target_type == "influxdb":
                    from .influxdb import InfluxDBTarget

                    target = InfluxDBTarget(self, "asab:metrics:{}".format(target))

                elif target_type == "http":
                    from .http import HTTPTarget

                    target = HTTPTarget(self, "asab:metrics:{}".format(target))

                else:
                    raise RuntimeError("Unknown target type {}".format(target_type))

                self.Targets.append(target)

        if Config.getboolean("asab:metrics", "native_metrics"):
            from .native import NativeMetrics

            self._native_svc = NativeMetrics(self.App, self)

    async def finalize(self, app):
        await self._on_flushing_event("finalize!")

    def clear(self):
        self.Metrics.clear()
        self.Storage.clear()

    def _flush_metrics(self):
        now = self.App.time()

        self.App.PubSub.publish("Metrics.flush!")
        for metric in self.Metrics:
            try:
                metric.flush(now)
            except Exception as e:
                L.exception(f"Exception during metric.flush()\n{e}")

        return now

    async def _on_flushing_event(self, event_type):
        if len(self.Metrics) == 0:
            return

        now = self._flush_metrics()

        pending = set()
        for target in self.Targets:
            pending.add(
                asyncio.ensure_future(target.process(self.Storage.Metrics, now))
            )

        while len(pending) > 0:
            done, pending = await asyncio.wait(
                pending, timeout=180.0, return_when=asyncio.ALL_COMPLETED
            )

    def _add_metric(
        self,
        metric: Metric,
        metric_name: str,
        tags=None,
        reset=None,
        help=None,
        unit=None,
    ):
        # Add global tags
        metric.StaticTags.update(self.Tags)
        metric.App = self.App

        # Add local static tags
        if tags is not None:
            for key, value in tags.items():
                # Check if every key and value is of type string. If not, try to convert it.
                assert isinstance(
                    key, str
                ), "Cannot add metrics tag: key '{}' is not a string.".format(key)
                assert isinstance(
                    value, str
                ), "Cannot add metrics tag for key '{}': value '{}' is not a string.".format(
                    key, value
                )
            metric.StaticTags.update(tags)

        metric._initialize_storage(
            self.Storage.add(
                metric_name,
                tags=metric.StaticTags.copy(),
                reset=reset,
                help=help,
                unit=unit,
            )
        )

        self.Metrics.append(metric)

    def create_gauge(
        self,
        metric_name: str,
        tags: dict = None,
        init_values: dict = None,
        help: str = None,
        unit: str = None,
    ):
        """
        The `create_gauge` function creates and returns a Gauge metric with specified parameters.

        Args:
                metric_name (str): The name of the metric you want to create.
                tags (dict): Dictionary where the keys represent the tag names and the values represent the tag values. It allows you
                        to categorize and filter metrics based on different dimensions or attributes.
                init_values (dict): The `init_values` parameter is used to set the initial value of the gauge
                        metric. Provide pairs of value names and their initial values. If not provided, the gauge
                        metric will be initialized with a value of 0.
                help (str): The "help" parameter is used to provide a description or explanation of the metric. It
                        can be used to provide additional information about what the metric measures or how it should be
                        interpreted.
                unit (str): The "unit" parameter is used to specify the unit of measurement for the metric. It
                        helps provide context and understanding for the metric being measured. For example, if the metric is
                        measuring the temperature, the unit could be "degrees Celsius".

        Returns:
                an instance of the Gauge class.

        Raises:
                AssertionError: `tags` dictionary has to be of type 'str': 'str'.
        """

        m = Gauge(init_values=init_values)
        self._add_metric(m, metric_name, tags=tags, help=help, unit=unit)
        return m

    def create_counter(
        self,
        metric_name: str,
        tags: dict = None,
        init_values: dict = None,
        reset: bool = True,
        help: str = None,
        unit: str = None,
        dynamic_tags: bool = False,
    ):
        """
        The function creates a counter metric with optional dynamic tags and adds it to a metric collection.

        Args:
                metric_name (str): The name of the metric you want to create.
                tags (dict): Dictionary where the keys represent the tag names and the values represent the tag values. It allows you
                        to categorize and filter metrics based on different dimensions or attributes.
                init_values (dict): The `init_values` parameter is used to set the initial value of the
                        metric. Provide pairs of value names and their initial values. If not provided, the
                        metric will be initialized with a value of 0.
                reset (bool): The "reset" parameter is a boolean value that determines whether the counter should
                        be reset to initial values every 60 seconds. Defaults to True
                help (str): The "help" parameter is used to provide a description or explanation of the metric. It
                        can be used to provide additional information about what the metric measures or how it should be
                        interpreted.
                unit (str): The "unit" parameter is used to specify the unit of measurement for the metric. It
                        helps provide context and understanding for the metric being measured. For example, if the metric is
                        measuring the temperature, the unit could be "degrees Celsius".
                dynamic_tags: Boolean flag. If set to True, the counter will be an instance of the
                        "CounterWithDynamicTags" class, which allows tags to be added or removed dynamically. Defaults to False

        Returns:
                the created counter object.

        Raises:
                AssertionError: `tags` dictionary has to be of type 'str': 'str'.
        """
        if dynamic_tags:
            m = CounterWithDynamicTags(init_values=init_values)
        else:
            m = Counter(init_values=init_values)
        self._add_metric(m, metric_name, tags=tags, reset=reset, help=help, unit=unit)
        return m

    def create_eps_counter(
        self,
        metric_name: str,
        tags: dict = None,
        init_values: dict = None,
        reset: bool = True,
        help: str = None,
        unit: str = None,
    ):
        """
        The function creates an EPSCounter object, and returns the object.

        Args:
                metric_name (str): The name of the metric you want to create.
                tags (dict): Dictionary where the keys represent the tag names and the values represent the tag values. It allows you
                        to categorize and filter metrics based on different dimensions or attributes.
                init_values (dict): The `init_values` parameter is used to set the initial value of the
                        metric. Provide pairs of value names and their initial values. If not provided, the
                        metric will be initialized with a value of 0.
                reset (bool): The "reset" parameter is a boolean value that determines whether the counter should
                        be reset to initial values every 60 seconds. Defaults to True
                help (str): The "help" parameter is used to provide a description or explanation of the metric. It
                        can be used to provide additional information about what the metric measures or how it should be
                        interpreted.
                unit (str): The "unit" parameter is used to specify the unit of measurement for the metric. It
                        helps provide context and understanding for the metric being measured. For example, if the metric is
                        measuring the temperature, the unit could be "degrees Celsius".

        Returns:
                an instance of the `EPSCounter` class.

        Raises:
                AssertionError: `tags` dictionary has to be of type 'str': 'str'.
        """
        m = EPSCounter(init_values=init_values)
        self._add_metric(m, metric_name, tags=tags, reset=reset, help=help, unit=unit)
        return m

    def create_duty_cycle(
        self,
        metric_name: str,
        tags: dict = None,
        init_values: dict = None,
        help: str = None,
        unit: str = None,
    ):
        """
        The function creates a duty cycle metric and returns the object.

        Args:
                metric_name (str): The name of the metric you want to create.
                tags (dict): Dictionary where the keys represent the tag names and the values represent the tag values. It allows you
                        to categorize and filter metrics based on different dimensions or attributes.
                init_values (dict): The `init_values` parameter is used to set the initial value of the
                        metric. Provide pairs of value names and their initial values. If not provided, the
                        metric will be initialized with a value of 0.
                help (str): The "help" parameter is used to provide a description or explanation of the metric. It
                        can be used to provide additional information about what the metric measures or how it should be
                        interpreted.
                unit (str): The "unit" parameter is used to specify the unit of measurement for the metric. It
                        helps provide context and understanding for the metric being measured. For example, if the metric is
                        measuring the temperature, the unit could be "degrees Celsius".

        Returns:
                an instance of the DutyCycle class.

        Raises:
                AssertionError: `tags` dictionary has to be of type 'str': 'str'.
        """
        m = DutyCycle(self.App, init_values=init_values)
        self._add_metric(m, metric_name, tags=tags, help=help, unit=unit)
        return m

    def create_aggregation_counter(
        self,
        metric_name,
        tags=None,
        init_values=None,
        reset: bool = True,
        aggregator=max,
        help=None,
        unit=None,
        dynamic_tags=False,
    ):
        """
        The function creates a counter metric with optional dynamic tags and adds it to a metric collection.

        Args:
                metric_name (str): The name of the metric you want to create.
                tags (dict): Dictionary where the keys represent the tag names and the values represent the tag values. It allows you
                        to categorize and filter metrics based on different dimensions or attributes.
                init_values (dict): The `init_values` parameter is used to set the initial value of the
                        metric. Provide pairs of value names and their initial values. If not provided, the
                        metric will be initialized with a value of 0.
                reset (bool): The "reset" parameter is a boolean value that determines whether the counter should
                        be reset to initial values every 60 seconds. Defaults to True
                help (str): The "help" parameter is used to provide a description or explanation of the metric. It
                        can be used to provide additional information about what the metric measures or how it should be
                        interpreted.
                unit (str): The "unit" parameter is used to specify the unit of measurement for the metric. It
                        helps provide context and understanding for the metric being measured. For example, if the metric is
                        measuring the temperature, the unit could be "degrees Celsius".
                dynamic_tags: Boolean flag. If set to True, the counter will be an instance of the
                        "AggregationCounterWithDynamicTags" class, which allows tags to be added or removed dynamically. Defaults to False

        Returns:
                the created counter object.

        Raises:
                AssertionError: `tags` dictionary has to be of type 'str': 'str'.
        """
        if dynamic_tags:
            m = AggregationCounterWithDynamicTags(
                init_values=init_values, aggregator=aggregator
            )
        else:
            m = AggregationCounter(init_values=init_values, aggregator=aggregator)
        self._add_metric(m, metric_name, tags=tags, reset=reset, help=help, unit=unit)
        return m

    def create_histogram(
        self,
        metric_name,
        buckets: list,
        tags=None,
        init_values=None,
        reset: bool = True,
        help=None,
        unit=None,
        dynamic_tags=False,
    ):
        """
        The function creates a histogram metric.

        Args:
                metric_name (str): The name of the metric you want to create.
                buckets (list): The "buckets" parameter is a list that specifies the boundaries for the histogram
                buckets. Each element in the list represents the upper bound of a bucket. For example, if the list
                is [10, 20, 30], it means that the histogram will have three buckets: one for values less than 10, second for values less than 20, third for values less than 30
                tags (dict): Dictionary where the keys represent the tag names and the values represent the tag values. It allows you
                        to categorize and filter metrics based on different dimensions or attributes.
                init_values (dict): The `init_values` parameter is used to set the initial value of the
                        metric. Provide pairs of value names and their initial values. If not provided, the
                        metric will be initialized with a value of 0.
                reset (bool): The "reset" parameter is a boolean value that determines whether the histogram should
                        be reset to initial values every 60 seconds. Defaults to True
                help (str): The "help" parameter is used to provide a description or explanation of the metric. It
                        can be used to provide additional information about what the metric measures or how it should be
                        interpreted.
                unit (str): The "unit" parameter is used to specify the unit of measurement for the metric. It
                        helps provide context and understanding for the metric being measured. For example, if the metric is
                        measuring the temperature, the unit could be "degrees Celsius".
                dynamic_tags: Boolean flag. If set to True, the counter will be an instance of the
                        "AggregationCounterWithDynamicTags" class, which allows tags to be added or removed dynamically. Defaults to False

        Returns:
                a histogram object

        Raises:
                AssertionError: `tags` dictionary has to be of type 'str': 'str'.
        """
        if dynamic_tags:
            m = HistogramWithDynamicTags(buckets=buckets, init_values=init_values)
        else:
            m = Histogram(buckets=buckets, init_values=init_values)
        self._add_metric(m, metric_name, tags=tags, reset=reset, help=help, unit=unit)
        return m



import mongoquery

import bspump.filter
import bspump.unittest


class ContentFilterOnHitPassOnly(bspump.filter.ContentFilter):
    def on_hit(self, context, event):
        return event

    def on_miss(self, context, event):
        return None


class TestContentFilter(bspump.unittest.ProcessorTestCase):
    def test_content_filter(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(bspump.filter.ContentFilter)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )

    def test_content_filter_with_query(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(bspump.filter.ContentFilter, query={"foo": "bar"})

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )

    def test_custom_content_filter_with_query(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(ContentFilterOnHitPassOnly, query={"foo": "bar"})

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"foo": "bar"}])

    def test_content_filter_with_invalid_query(self):
        with self.assertRaises(mongoquery.QueryError):
            self.set_up_processor(bspump.filter.ContentFilter, query={"$foo": 2})

        with self.assertRaises(TypeError):
            self.set_up_processor(
                bspump.filter.ContentFilter, query={"foo": {"$in": None}}
            )



import xxhash

from ..value.valueexpr import VALUE
from ...abc import Expression

###


class HASH(Expression):
    Attributes = {
        "What": ["*"],
        "Seed": [
            "si64",
            "si8",
            "si16",
            "si32",  # Those are not yet supported 'si128', 'si256',
            "ui8",
            "ui16",
            "ui32",
            "ui64",  # Those are not yet supported 'ui128', 'ui256',
        ],
        "Type": ["str"],  # And must be evaluated in compile type
    }

    def __init__(self, app, *, arg_what, arg_seed: int = 0, arg_type="xxhash64"):
        super().__init__(app)

        if isinstance(arg_type, Expression):
            self.Type = arg_type
        else:
            self.Type = VALUE(app, value=arg_type)
        arg_type = self.Type(None, None)
        assert isinstance(arg_type, str)
        assert arg_type in ("xxhash64")

        if isinstance(arg_what, Expression):
            self.What = arg_what
        else:
            self.What = VALUE(app, value=arg_what)

        if isinstance(arg_seed, Expression):
            self.Seed = arg_seed
        else:
            self.Seed = VALUE(app, value=arg_seed)

    def initialize(self):
        outlet_type = self.What.get_outlet_type()
        self.Method = get_hash_method(self.Type(None, None), outlet_type)

    def __call__(self, context, event, *args, **kwargs):
        what = self.What(context, event, *args, **kwargs)
        seed = self.Seed(context, event, *args, **kwargs)
        return self.Method(what, seed)

    def get_outlet_type(self):
        return "ui64"


def get_hash_method(hash_type, input_type):
    if hash_type == "xxhash64":
        if input_type in (
            "int",
            "si64",
            "si8",
            "si16",
            "si32",
            "si128",
            "si256",
            "ui8",
            "ui16",
            "ui32",
            "ui64",
            "ui128",
            "ui256",
        ):
            # For integer, use the alternative hashing functions
            return hash_xxhash64_int

        elif input_type in ("str"):
            return hash_xxhash64_str

    raise NotImplementedError(
        "Hashing for '{}' and '{}' is not yet implemented".format(hash_type, input_type)
    )


def hash_xxhash64_str(what, seed):
    return xxhash.xxh64(what, seed=seed).intdigest()


def hash_xxhash64_int(what, seed):
    return (what ^ seed) & 0xFF_FF_FF_FF_FF_FF_FF_FF  # Limit to 64bits



from ...abc import Expression

from ..value.valueexpr import VALUE
from ..value.eventexpr import EVENT


class IN(Expression):
    """
    Checks if expression is of given list.
    """

    Attributes = {
        "What": ["*"],  # TODO: 'What' is different from list, set and dict
        "Where": ["list", "set", "dict"],
    }

    Category = "Compare"

    def __init__(self, app, *, arg_what, arg_where):
        super().__init__(app)

        if isinstance(arg_what, Expression):
            self.What = arg_what
        else:
            self.What = VALUE(app, value=arg_what)

        if isinstance(arg_where, Expression):
            self.Where = arg_where
        else:
            self.Where = VALUE(app, value=arg_where)

    def optimize(self):
        if isinstance(self.Where, VALUE):
            if self.Where.get_outlet_type() == "list":
                return IN_optimized_list_where(self)

            if self.Where.get_outlet_type() == "set":
                return IN_optimized_set_where(self)

        if isinstance(self.What, VALUE) and isinstance(self.Where, EVENT):
            return IN_optimized_EVENT_VALUE(self)

        return None

    def get_items_inlet_type(self):
        # Find the first usable type in the items
        for item in self.Where.Value:
            if isinstance(item, str):
                return "str"

            elif isinstance(item, int):
                return "int"

            elif isinstance(item, Expression):
                outlet_type = item.get_outlet_type()
                if outlet_type not in frozenset(["^"]):
                    return outlet_type

        raise NotImplementedError("Cannot decide on items inlet type '{}'".format(self))

    def consult_inlet_type(self, key, child):
        return self.get_items_inlet_type()

    def __call__(self, context, event, *args, **kwargs):
        return self.What(context, event, *args, **kwargs) in self.Where(
            context, event, *args, **kwargs
        )

    def get_outlet_type(self):
        return bool.__name__


class IN_optimized_list_where(IN):
    def __init__(self, orig):
        super().__init__(orig.App, arg_what=orig.What, arg_where=orig.Where)

        self._where_value = frozenset(self.Where({}, {}))

    def optimize(self):
        # This is to prevent re-optimising the class
        return None

    def __call__(self, context, event, *args, **kwargs):
        return self.What(context, event, *args, **kwargs) in self._where_value


class IN_optimized_set_where(IN):
    def __init__(self, orig):
        super().__init__(orig.App, arg_what=orig.What, arg_where=orig.Where)

        self._where_value = self.Where({}, {})

    def optimize(self):
        # This is to prevent re-optimising the class
        return None

    def __call__(self, context, event, *args, **kwargs):
        return self.What(context, event, *args, **kwargs) in self._where_value


class IN_optimized_EVENT_VALUE(IN):
    def __init__(self, orig):
        super().__init__(orig.App, arg_what=orig.What, arg_where=orig.Where)

        self._what_value = self.What({}, {})

    def optimize(self):
        # This is to prevent re-optimising the class
        return None

    def __call__(self, context, event, *args, **kwargs):
        return self._what_value in event



import logging

import secrets
import hashlib
import base64

import orjson

from ..abc.processor import Processor

###

L = logging.getLogger(__name__)

###


class IntegrityEnricher(Processor):
    """
    IntegrityEnricher is a enricher processor, which enriches JSON data
    by hashed events.

    Supported algorithms for cryptographic signing, default is SHA256
    'SHA256', 'dsaEncryption', 'MD4', 'sha256', 'sha3_512', 'DSA', 'sha3_256', 'sha3_384', 'SHA512', 'md5', 'SHA224',
    'MD5', 'sha', 'whirlpool', 'ripemd160', 'SHA384', 'ecdsa-with-SHA1', 'RIPEMD160', 'sha1', 'blake2s', 'shake_128',
    'blake2b', 'sha512', 'sha224', 'md4', 'SHA', 'dsaWithSHA', 'sha384', 'sha3_224', 'shake_256', 'DSA-SHA', 'SHA1'
    """

    ConfigDefaults = {
        "algorithm": "SHA256",
        "hash_key": "_id",
        "prev_hash_key": "_prev_id",
        "salt_length": 3,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        self.Algorithm = self.Config["algorithm"]
        self.HashKey = self.Config["hash_key"]
        self.PrevHashKey = self.Config["prev_hash_key"]
        self.SaltLength = int(self.Config["salt_length"])
        self.PreviousHash = None

    def process(self, context, event):
        # Check that the event is a dictionary
        assert isinstance(event, dict)

        # Check if hash / previous hash already present in event and if so, delete it from event
        event.pop(self.HashKey, None)

        # Salt event - to ensure that events are not going to be the same after hash
        event["_s"] = secrets.token_urlsafe(self.SaltLength)

        # Set previous hash
        if self.PreviousHash is not None:
            event[self.PrevHashKey] = self.PreviousHash
        else:
            event.pop(self.PrevHashKey, None)

        # Hash event using key, value, key, value ... sequence
        h = hashlib.new(self.Algorithm)
        h.update(orjson.dumps(event, option=orjson.OPT_SORT_KEYS))
        hash_base64 = h.hexdigest()

        # Store the hash as base64 string
        event[self.HashKey] = hash_base64

        # Actual hash will become previous hash in the next iteration
        self.PreviousHash = hash_base64

        return event


def orjson_default(obj):
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode("ascii")
    raise TypeError



import logging

from ..config import Config
from .service import LibraryService

#

L = logging.getLogger(__name__)

#

Config.add_defaults({"library": {"providers": "./library"}})

__all__ = [
    "LibraryService",
]



import logging

import mongoquery

from ..abc.processor import Processor

###

L = logging.getLogger(__name__)

###


class ContentFilter(Processor):
    """
    This is processor implenting a simple attribute filter.
    The query is constructed in mongodb manner.
    NB:
    Only the "/pattern/<options>" syntax is supported for $regex.
    As a consequence, $options isn’t supported.
    $text hasn’t been implemented.
    Due to the pure python nature of mongoquery, $where isn’t supported.
    The Geospatial operators $geoIntersects, $geoWithin, $nearSphere, and $near are not implemented.
    Projection operators $`, $elemMatch, $meta,
    and $slice are not implemented (only querying is implemented)
    $type is limited to recognising generic python types,
    it won’t look into recognising the format of
    the data (for instance, it doesn’t check Object ID’s format, only that they are strings).
    (from https://pypi.org/project/mongoquery/)
    """

    def __init__(self, app, pipeline, query={}, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        # Check if the query is correctly implemented
        try:
            self.Query = mongoquery.Query(query)
            self.Query.match({})
        except mongoquery.QueryError:
            L.warning("Incorrect query")
            raise

    def on_hit(self, context, event):
        """
        This function tranforms the event, if it
        matched the query.
        """
        return event

    def on_miss(self, context, event):
        """
        This function tranforms the event, if it did not
        matched the query.
        """
        return event

    def process(self, context, event):
        matched = self.Query.match(event)
        if matched:
            new_event = self.on_hit(context, event)
        else:
            new_event = self.on_miss(context, event)

        return new_event



import logging

from ..abc.sink import Sink

#

L = logging.getLogger(__name__)

#


class InfluxDBSink(Sink):
    """
            Description: InfluxDBSink is a sink processor, that stores the event into an InfluxDB database
            specified in the InfluxDBConnection object.

    .. code:: python

            class SamplePipeline(bspump.Pipeline):

                    def __init__(self, app, pipeline_id):
                            super().__init__(app, pipeline_id)
                            self.build(
                                    bspump.socket.TCPStreamSource(app, self, config={'port': 7000}),
                                    bspump.influxdb.InfluxDBSink(app, self, "InfluxConnection1")
                            )

    """

    def __init__(self, app, pipeline, connection, id=None, config=None):
        """
        Description:

        **Parameters**

        app :

        pipeline :

        connection :

        id : ID, default = None

        config : str,JSON, default = None

        """
        super().__init__(app, pipeline, id=id, config=config)
        self._connection = pipeline.locate_connection(app, connection)

        app.PubSub.subscribe("InfluxDBConnection.pause!", self._connection_throttle)
        app.PubSub.subscribe("InfluxDBConnection.unpause!", self._connection_throttle)

    # TODO: Restructure data: { "measurement": "location", "tag_set": "location=us-midwest", "field_set": "temperature=82", "timestamp": 1465839830100400200 }
    def process(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                Information with timestamp.


        """

        if isinstance(event, tuple):
            measurement, tag_set, field_set, timestamp = event
            wire_line = "{},{} {} {}\n".format(
                measurement, tag_set, field_set, int(timestamp * 1e9)
            )

        elif isinstance(event, bytes):
            wire_line = event.decode("utf-8")
            if wire_line[-1:] != "\n":
                wire_line += "\n"

        elif isinstance(event, str):
            wire_line = event
            if wire_line[-1:] != "\n":
                wire_line += "\n"

        else:
            raise RuntimeError("Incorrect format")

        # Passing the processed event to the connection
        self._connection.consume(wire_line)

    def _connection_throttle(self, event_name, connection):
        if connection != self._connection:
            return

        if event_name == "InfluxDBConnection.pause!":
            self.Pipeline.throttle(self, True)
        elif event_name == "InfluxDBConnection.unpause!":
            self.Pipeline.throttle(self, False)
        else:
            raise RuntimeError("Unexpected event name '{}'".format(event_name))



import asyncio
import logging

import aiomysql
import aiomysql.utils

import pymysql
import pymysql.cursors
import pymysql.err

from ..abc.connection import Connection
from .convertors import convertors

#

L = logging.getLogger(__name__)

#


class MySQLConnection(Connection):
    """
            MySQLConnection is a top BitSwan object, that is used to connect the application to an external MySQL database.
            MySQLConnection is thus used in MySQL-related sources, sinks, processors and lookups such as MySQLLookup.

            MySQLConnection is built on top of aiomysql library and utilizes its functions: https://aiomysql.readthedocs.io/en/latest/

            The following code illustrates how to create and register the MySQL connection inside the application object.
            The connection's ID is set to "MySQLConnection" (the second argument of the connection's constructor),
            which is also then used as a section name in configuration files.

    .. code:: python

            app = bspump.BSPumpApplication()
            svc = app.get_service("bspump.PumpService")
            svc.add_connection(
                    bspump.mysql.MySQLConnection(app, "MySQLConnection")
            )

            In order to locate MySQLConnection in the constructor method of sources, sinks, processors etc., the following code can be used:

    .. code:: python

            svc = app.get_service("bspump.PumpService")
            connection = svc.locate_connection("MySQLConnection")

    """

    ConfigDefaults = {
        "host": "localhost",
        "port": 3306,
        "user": "",
        "password": "",
        "db": "",
        "connect_timeout": 10,
        "reconnect_delay": 5.0,
        "output_queue_max_size": 100,
        "max_bulk_size": 5,
        "autocommit": True,
        "connection_pool_minsize": 1,
        "connection_pool_maxsize": 10,
    }

    RetryErrors = frozenset(
        [
            1047,
            1053,
            1077,
            1078,
            1079,
            1080,
            2003,
            2006,
            2012,
            2013,
            1152,
            1205,
            1213,
            1223,
        ]
    )

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.ConnectionEvent = asyncio.Event()
        self.ConnectionEvent.clear()

        self.PubSub = app.PubSub
        self.Loop = app.Loop

        self._host = self.Config["host"]
        self._port = int(self.Config["port"])
        self._user = self.Config["user"]
        self._password = self.Config["password"]
        self._connect_timeout = int(self.Config["connect_timeout"])
        self._db = self.Config["db"]
        self._reconnect_delay = float(self.Config["reconnect_delay"])
        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self._max_bulk_size = int(self.Config["max_bulk_size"])
        self._autocommit = self.Config.getboolean("autocommit")

        self._conn_future = None
        self._conn_pool = None
        self._conn_sync = None
        self._connection_request = False
        self._pause = False
        self._throttled_by_error = False

        # Size of connection pool
        self.ConnectionPoolMinSize = self.Config["connection_pool_minsize"]
        self.ConnectionPoolMaxSize = self.Config["connection_pool_maxsize"]

        # Subscription
        self._on_health_check("connection.open!")
        app.PubSub.subscribe("Application.stop!", self._on_application_stop)
        app.PubSub.subscribe("Application.tick!", self._on_health_check)
        app.PubSub.subscribe("MySQLConnection.pause!", self._on_pause)
        app.PubSub.subscribe("MySQLConnection.unpause!", self._on_unpause)

        self._output_queue = asyncio.Queue()
        self._bulks = {}  # We have a "bulk" per query

    def _on_pause(self, event_type, connection):
        self._pause = True

    def _on_unpause(self, event_type, connection):
        self._pause = False

    def _flush(self):
        for query in self._bulks.keys():
            # Break if throttling was requested during the flush,
            # so that put_nowait doesn't raise
            if self._pause:
                break

            self._flush_bulk(query)

    def _flush_bulk(self, query):
        # Enqueue and throttle if needed
        self._output_queue.put_nowait((query, self._bulks[query]))
        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("MySQLConnection.pause!", self)

        # Reset bulk
        self._bulks[query] = []

    def _on_application_stop(self, message_type, counter):
        self._flush()
        self._output_queue.put_nowait((None, None))

    def _on_health_check(self, message_type):
        if self._conn_future is not None:
            # Connection future exists

            if not self._conn_future.done():
                # Connection future didn't result yet
                # No sanitization needed
                return

            try:
                self._conn_future.result()
            except Exception:
                # Connection future threw an error
                L.exception("Unexpected connection future error")

            # Connection future already resulted (with or without exception)
            self._conn_future = None

        assert self._conn_future is None

        self._conn_future = asyncio.ensure_future(
            self._async_connection(),
        )

        self._sync_connection()

    async def _async_connection(self):
        # Erase the previous connections
        if self._conn_pool is not None:
            try:
                self._conn_pool.close()
                await self._conn_pool.wait_closed()
            except (
                pymysql.err.InternalError,
                pymysql.err.ProgrammingError,
                pymysql.err.OperationalError,
            ):
                pass
            self._conn_pool = None

        try:
            async with aiomysql.create_pool(
                host=self._host,
                port=self._port,
                user=self._user,
                password=self._password,
                db=self._db,
                conv=convertors,
                connect_timeout=self._connect_timeout,
                autocommit=self._autocommit,
                minsize=self.ConnectionPoolMinSize,
                maxsize=self.ConnectionPoolMaxSize,
                loop=self.Loop,
            ) as pool:
                self._conn_pool = pool
                self.ConnectionEvent.set()

                # Ensures that all connections are utilized
                loaders = []
                for index in range(0, self.ConnectionPoolMaxSize):
                    loaders.append(asyncio.ensure_future(self._loader()))
                await asyncio.wait(loaders, return_when=asyncio.ALL_COMPLETED)

        except (
            pymysql.err.InternalError,
            pymysql.err.ProgrammingError,
            pymysql.err.OperationalError,
        ) as e:
            if e.args[0] in self.RetryErrors:
                L.warn(
                    "Recoverable error '{}' occurred in MySQLConnection. Retrying.".format(
                        e.args[0]
                    )
                )
                self._conn_future = None
                return
            L.exception("Unexpected MySQL connection error")
            raise e
        except BaseException:
            L.exception("Unexpected MySQL connection error")
            raise

    def _sync_connection(self):
        # Close the connection first
        if self._conn_sync is not None:
            try:
                self._conn_sync.close()
            except (
                pymysql.err.InternalError,
                pymysql.err.ProgrammingError,
                pymysql.err.OperationalError,
            ):
                pass
            self._conn_sync = None

        try:
            connection = pymysql.connect(
                host=self._host,
                port=self._port,
                user=self._user,
                password=self._password,
                database=self._db,
                conv=convertors,
                autocommit=self._autocommit,
                connect_timeout=self._connect_timeout,
            )
            self._conn_sync = connection
        except (
            pymysql.err.InternalError,
            pymysql.err.ProgrammingError,
            pymysql.err.OperationalError,
        ) as e:
            if e.args[0] in self.RetryErrors:
                L.warn(
                    "Recoverable error '{}' occurred in MySQLConnection. Retrying.".format(
                        e.args[0]
                    )
                )
                return
            L.exception("Unexpected MySQL connection error")
            raise e
        except BaseException:
            L.exception("Unexpected MySQL connection error")
            raise

    def acquire_connection(self) -> aiomysql.utils._PoolAcquireContextManager:
        """
                Acquire asynchronous database connection

                Use with `async with` statement

        .. code-block:: python

                async with self.Connection.acquire_connection() as connection:
                        async with connection.cursor() as cursor:
                                await cursor.execute(query)

                :return: Asynchronous Context Manager
        """
        assert self._conn_pool is not None
        return self._conn_pool.acquire()

    def acquire_sync_cursor(self) -> pymysql.cursors.DictCursor:
        """
                Acquire synchronous database cursor

                Use with `with` statement

        .. code-block:: python

                with self.Connection.acquire_sync_cursor() as cursor:
                        await cursor.execute(query)

                :return: Context Manager
        """
        assert self._conn_sync is not None
        return pymysql.cursors.DictCursor(self._conn_sync)

    def consume(self, query, data):
        # Create a bulk for this query if doesn't yet exist
        if query not in self._bulks:
            self._bulks[query] = []

        # Add data to the query's bulk
        self._bulks[query].append(data)

        # Flush on _max_bulk_size
        if len(self._bulks[query]) >= self._max_bulk_size:
            self._flush_bulk(query)

    async def _loader(self):
        while True:
            query, data = await self._output_queue.get()

            if query is None:
                break

            if self._output_queue.qsize() == self._output_queue_max_size - 1:
                self.PubSub.publish(
                    "MySQLConnection.unpause!", self, asynchronously=True
                )

            async with self.acquire_connection() as connection:
                async with connection.cursor() as cursor:
                    try:
                        await cursor.executemany(query, data)
                        await connection.commit()
                        if self._throttled_by_error:
                            self.Pipeline.throttle(self, False)
                            self._throttled_by_error = False
                    except (
                        pymysql.err.InternalError,
                        pymysql.err.ProgrammingError,
                        pymysql.err.OperationalError,
                    ) as e:
                        if e.args[0] in self.RetryErrors:
                            L.warn(
                                "Recoverable error '{}' occurred in MySQLConnection. Retrying.".format(
                                    e.args[0]
                                )
                            )
                            self._output_queue.put_nowait((query, data))
                            self.Pipeline.throttle(self, True)
                            self._throttled_by_error = True
                        raise e



from bspump.declarative.abc import Expression
from ..value.valueexpr import VALUE


class MAP(Expression):
    """
    Checks if a given value is in the provided map in items.

    Usage:

    !MAP
    value: !ITEM EVENT potatoes
    map:
            7: only seven
            20: twenty
            12: twelve
            10: enough
    else:
            nothing found
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Default": ["*"],  # TODO: This ...
    }

    def __init__(self, app, *, arg_what, arg_in, arg_else=None):
        super().__init__(app)
        self.Value = arg_what
        self.Default = arg_else

        if not isinstance(arg_else, Expression):
            self.Default = VALUE(app, value=arg_else)
        else:
            self.Default = arg_else

        self.Map = arg_in
        self.MapNormalized = dict()
        self.OutletType = None  # Will be determined in `initialize()`

    def set(self, key, value):
        setattr(self, key, value)

        if "Map" in key:
            self.MapNormalized[key[3:]] = value

    def initialize(self):
        for key, value in self.Map.items():
            if not isinstance(value, Expression):
                value = VALUE(self.App, value=value)

            attr_name = "Map{}".format(key)
            setattr(self, attr_name, value)
            if self.OutletType is None:
                self.OutletType = value.get_outlet_type()
            self.Attributes[attr_name] = self.OutletType

            self.MapNormalized[key] = value

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        try:
            item = self.MapNormalized[value]
        except KeyError:
            return self.Default(context, event, *args, **kwargs)
        return item(context, event, *args, **kwargs)



from pymysql.converters import encoders as convertors
import numpy as np

""" Those methods extends pymysql convertor for numpy datatypes """


def convert_numpy_int(value, mapping=None):
    return str(value)


def convert_numpy_float(value, mapping=None):
    s = repr(value)
    if s in ("inf", "nan"):
        raise ValueError("%s can not be used with MySQL" % s)
    if "e" not in s:
        s += "e0"
    return s


convertors[np.int8] = convert_numpy_int
convertors[np.int16] = convert_numpy_int
convertors[np.int32] = convert_numpy_int
convertors[np.int64] = convert_numpy_int

convertors[np.uint8] = convert_numpy_int
convertors[np.uint16] = convert_numpy_int
convertors[np.uint32] = convert_numpy_int
convertors[np.uint64] = convert_numpy_int

convertors[np.float16] = convert_numpy_float
convertors[np.float32] = convert_numpy_float
convertors[np.float64] = convert_numpy_float



import asyncio
import logging
import random
import re

import aiohttp

from ..abc.connection import Connection

#

L = logging.getLogger(__name__)

#


class ElasticSearchBulk(object):
    """
    Description:

    """

    def __init__(self, connection, index, max_size):
        """
        Initializes the variables

        **Parameters**

        connection : Connection
                        Name of the Connection.

        index : str
                        ???

        max_size : int
                        Maximal size of bulks.

        """
        self.Index = index
        self.Aging = 0
        self.Capacity = max_size
        self.Items = []
        self.InsertMetric = connection.InsertMetric
        self.FailLogMaxSize = connection.FailLogMaxSize
        self.FilterPath = connection.FilterPath

    def consume(self, data_feeder_generator):
        """
        Appends all items in data_feeder_generator to Items list. Consumer also resets Aging and Capacity.

        **Parameters**

        data_feeder_generator : list
                        list of our data that will be passed to a generator and later Uploaded to ElasticSearch.

        :return: self.Capacity <= 0

        """
        for item in data_feeder_generator:
            self.Items.append(item)
            self.Capacity -= len(item)

        self.Aging = 0
        return self.Capacity <= 0

    async def _get_data_from_items(self):
        """
        Description:

        :return:
        """
        for item in self.Items:
            yield item

    async def upload(self, url, session, timeout):
        """
        Uploads data to Elastic Search.

        **Parameters**

        url : string
                        Uses URL from config to connect to ElasticSearch Rest API.

        session : ?
                        ?

        timeout : int
                        uses timout value from config. Value of time for how long we want to be connected to ElasticSearch.

        :return: ?

        |

        """
        items_count = len(self.Items)
        if items_count == 0:
            return

        url = url + "{}/_bulk?filter_path={}".format(self.Index, self.FilterPath)

        try:
            resp = await session.post(
                url,
                data=self._get_data_from_items(),
                headers={"Content-Type": "application/json"},
                timeout=timeout,
            )
        except OSError as e:
            # This means that there was a hard error such as network or DNS failure
            # Likely no communication took place with ElasticSearch
            L.warn("{}".format(e))
            return False

        # Obtain the response from ElasticSearch, which should always be a json
        try:
            resp_body = await resp.json()
        except Exception as e:
            # We received something else than JSON, that's bad
            # Let's assume that the bulk did not reach ElasticSearch
            L.warn("{}".format(e))
            return False

        if resp.status == 200:
            # Check that all documents were successfully inserted to ElasticSearch
            # If there are no error messages, we are done here
            if not resp_body.get("errors", False):
                self.InsertMetric.add("ok", items_count)
                return True

            # Some of the documents were not inserted properly,
            # usually because of attributes mismatch, see:
            # https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html

            response_items = resp_body.get("items")
            self.partial_error_callback(response_items)

            # When the log handling is not precise,
            # the iteration only happens on error items natively
            # Log first 20 errors
            counter = 0
            for response_item in response_items:
                if "error" not in response_item.get("index", ""):
                    continue

                if counter < self.FailLogMaxSize:
                    L.error(
                        "Failed to insert document into ElasticSearch: '{}'".format(
                            str(response_item)
                        )
                    )

                counter += 1

            # Show remaining log messages
            if counter > self.FailLogMaxSize:
                L.error(
                    "Failed to insert document into ElasticSearch: '{}' more insertions of documents failed".format(
                        counter - self.FailLogMaxSize
                    )
                )

            # Insert metrics
            self.InsertMetric.add("fail", counter)
            self.InsertMetric.add("ok", items_count - counter)

        else:
            # The major ElasticSearch error occurred while inserting documents, response was not 200
            self.InsertMetric.add("fail", items_count)
            L.error(
                "Failed to insert document into ElasticSearch status:{} body:{}".format(
                    resp.status, resp_body
                )
            )
            return self.full_error_callback(self.Items, resp.status)

        return True

    def partial_error_callback(self, response_items):
        """
        Description: When an upload to ElasticSearch fails for error items (document could not be inserted),
        this callback is called.

        **Parameters**

        response_items :

        :param response_items: list with dict items: {"index": {"_id": ..., "error": ...}}

        :return:
        """
        L.error(
            "Failed to insert items in the elasticsearch: {}".format(
                response_items[:10]
            )
        )

    def full_error_callback(self, bulk_items, return_code):
        """
        Description: When an upload to ElasticSearch fails b/c of ElasticSearch error,
        this callback is called.

        **Parameters**

        bulk_items : list
                        list with tuple items: (_id, data)

        return_code :
                        ElasticSearch return code

        :return: False if the bulk is to be resumbitted again

        |

        """
        return False


class ElasticSearchConnection(Connection):
    """
    Description:

    **Sample Config**

    url : ''http'://{ip/localhost}:{port}'
                    URL of the source. Could be multi-URL. Each URL should be separated by ';' to a node in ElasticSearch cluster.

    username : 'string' , default = ' '
                    Used when authentication is required

    password : 'string', default = ' '
                    Used when authentication is required

    loader_per_url : int, default = 4
                    Number of parallel loaders per URL.

    output_queue_max_size : int, default = 10
                    Maximum queue size.

    bulk_out_max_size : ? * ? * ?, default = 12 * 1024 * 1024
                    ??

    timeout : int, default = 300
                    Timout value.

    fail_log_max_size : int, default =  20
                    Maximum size of failed log messages.

    precise_error_handling : bool, default = False
                    If True all Errors will be logged, If false soft errors will be omitted in the Logs.



    """

    ConfigDefaults = {
        "url": "http://localhost:9200/",
        # Could be multi-URL. Each URL should be separated by ';' to a node in ElasticSearch cluster
        "username": "",
        "password": "",
        "loader_per_url": 4,  # Number of parallel loaders per URL
        "output_queue_max_size": 10,
        "bulk_out_max_size": 12 * 1024 * 1024,
        "timeout": 300,
        "fail_log_max_size": 20,
        "precise_error_handling": False,
    }

    def __init__(self, app, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                        Name of the Application

        id : ID, default= None
                        ID

        config : JSON or dict, default= None
                        configuration file with additional information for the methods.

        """
        super().__init__(app, id=id, config=config)

        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self._output_queue = asyncio.Queue()

        username = self.Config.get("username")
        password = self.Config.get("password")

        if username == "":
            self._auth = None
        else:
            self._auth = aiohttp.BasicAuth(login=username, password=password)

        # Contains URLs of each node in the cluster
        self.node_urls = []
        # for url in self.Config['url'].split(';'):
        for url in re.split(r"\s+", self.Config["url"]):
            url = url.strip()
            if len(url) == 0:
                continue
            if url[-1] != "/":
                url += "/"
            self.node_urls.append(url)

        self._loader_per_url = int(self.Config["loader_per_url"])

        self._bulk_out_max_size = int(self.Config["bulk_out_max_size"])
        self._bulks = {}

        self._timeout = float(self.Config["timeout"])
        self._started = True

        self.Loop = app.Loop

        self.PubSub = app.PubSub
        self.PubSub.subscribe("Application.run!", self._start)
        self.PubSub.subscribe("Application.exit!", self._on_exit)

        self._futures = []
        for url in self.node_urls:
            for i in range(self._loader_per_url):
                self._futures.append((url, None))

        self.FailLogMaxSize = int(self.Config["fail_log_max_size"])

        # Precise error handling
        if self.Config.getboolean("precise_error_handling"):
            self.FilterPath = "errors,took,items.*.error,items.*._id"
        else:
            self.FilterPath = "errors,took,items.*.error"

        # Create metrics counters
        metrics_service = app.get_service("asab.MetricsService")
        self.InsertMetric = metrics_service.create_counter(
            "elasticsearch.insert",
            init_values={
                "ok": 0,
                "fail": 0,
            },
        )
        self.QueueMetric = metrics_service.create_gauge(
            "elasticsearch.outputqueue",
            init_values={
                "size": 0,
            },
        )

    def get_url(self):
        """
        :return: list of URLS of nodes connected to the cluster

        """
        return random.choice(self.node_urls)

    def get_session(self):
        """
        Returns current Client Session Authentication and Loop

        :return: aiohttp.ClientSession(auth=self._auth, loop=self.Loop)

        :return:
        """
        return aiohttp.ClientSession(auth=self._auth)

    def consume(self, index, data_feeder_generator, bulk_class=ElasticSearchBulk):
        """
        Checks the content of data_feeder_generator and bulk and if There is data to be send it calls enqueue method.

        **Parameters**

        index :

        data_feeder_generator :

        bulk_class=ElasticSearchBulk :
                        creates a instance of the ElasticSearchBulk class

        """
        if data_feeder_generator is None:
            return

        bulk = self._bulks.get(index)
        if bulk is None:
            bulk = bulk_class(self, index, self._bulk_out_max_size)
            self._bulks[index] = bulk

        if bulk.consume(data_feeder_generator):
            # Bulk is ready, schedule to be send
            del self._bulks[index]
            self.enqueue(bulk)

    def _start(self, event_name):
        """
        Description:

        :return:
        """
        self.PubSub.subscribe("Application.tick!", self._on_tick)
        self._on_tick("simulated!")

    async def _on_exit(self, event_name):
        """
        Description:

        :return:
        """
        # Wait till the queue is empty
        self.flush(forced=True)
        while self._output_queue.qsize() > 0:
            self.flush(forced=True)
            await asyncio.sleep(1)
            if self._output_queue.qsize() > 0:
                L.warn(
                    "Still have {} bulk in output queue".format(
                        self._output_queue.qsize()
                    )
                )

        self._started = False

        # Wait till the _loader() terminates (one after another)
        pending = [item[1] for item in self._futures]
        self._futures = []
        while len(pending) > 0:
            # By sending None via queue, we signalize end of life
            await self._output_queue.put(None)
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )

    def _on_tick(self, event_name):
        """
        Description:

        :return:
        """
        self.QueueMetric.set("size", int(self._output_queue.qsize()))

        for i in range(len(self._futures)):
            # 1) Check for exited futures
            url, future = self._futures[i]
            if future is not None and future.done():
                # Ups, _loader() task crashed during runtime, we need to restart it
                try:
                    future.result()
                    if self._started:
                        L.error("ElasticSearch issue detected, will retry shortly")
                except Exception as e:
                    L.exception(
                        f"ElasticSearch issue detected '{e}', will retry shortly"
                    )

                self._futures[i] = (url, None)

            # 2) Start _loader() futures that are exitted
            if self._started:
                url, future = self._futures[i]
                if future is None:
                    future = asyncio.ensure_future(self._loader(url))
                    self._futures[i] = (url, future)

        self.flush()

    def flush(self, forced=False):
        """
        It goes through the list of bulks and calls enqueue for each of them.

        **Parameters**

        forced : bool, default = False

        """
        aged = []

        for index, bulk in self._bulks.items():
            if bulk is None:
                continue

            bulk.Aging += 1

            if (bulk.Aging >= 2) or forced:
                aged.append(index)

        for index in aged:
            bulk = self._bulks.pop(index)
            self.enqueue(bulk)

    def enqueue(self, bulk):
        """
        Properly enqueue the bulk.

        **Parameters**

        bulk :

        """
        self._output_queue.put_nowait(bulk)

        # Signalize need for throttling
        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("ElasticSearchConnection.pause!", self)

    async def _loader(self, url):
        """
        Description:

        :return:
        """
        async with self.get_session() as session:
            # Preflight check
            try:
                async with session.get(url + "_cluster/health") as resp:
                    await resp.json()
                    if resp.status != 200:
                        L.error(
                            "Cluster is not ready", struct_data={"status": resp.status}
                        )
                        await asyncio.sleep(5)  # Throttle a bit before next try
                        return

            except aiohttp.client_exceptions.ServerDisconnectedError:
                L.error("Cluster is not ready, server disconnected or not ready")
                await asyncio.sleep(5)  # Throttle a bit before next try
                return

            except OSError as e:
                L.error("{}, cluster is not ready".format(e))
                await asyncio.sleep(5)  # Throttle a bit before next try
                return

            except aiohttp.client_exceptions.ContentTypeError as e:
                L.error("Failed communication {}".format(e))
                await asyncio.sleep(20)  # Throttle a lot before next try
                return

            except GeneratorExit as e:
                L.info("Generator exited {}".format(e))
                return

            # Push bulks into the ElasticSearch
            while self._started:
                bulk = await self._output_queue.get()
                if bulk is None:
                    break

                if self._output_queue.qsize() == self._output_queue_max_size - 1:
                    self.PubSub.publish("ElasticSearchConnection.unpause!", self)

                sucess = await bulk.upload(url, session, self._timeout)
                if not sucess:
                    # Requeue the bulk for another delivery attempt to ES
                    self.enqueue(bulk)
                    await asyncio.sleep(5)  # Throttle a bit before next try
                    break  # Exit the loader (new will be started automatically)

                # Make sure the memory is emptied
                bulk.Items = []



from .processor import ProcessorBase


class Generator(ProcessorBase):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                        Name of the Application.

        pipeline : Pipeline
                        Name of the Pipeline.

        id : str, default = None
                        ID

        config : JSON, defualt = None
                        configuration file containing additional information.

        """
        super().__init__(app, pipeline, id, config)
        # The correct depth is later set by the pipeline
        self.PipelineDepth = None

    def set_depth(self, depth):
        """
        Description:

        **Parameters**

        depth : int


        """
        assert self.PipelineDepth is None
        self.PipelineDepth = depth

    def process(self, context, event):
        """
        Description:


        **Parameters**

        context :

        event : any data type
                        information of any data type with timestamp.


        """
        self.Pipeline.ensure_future(
            self.generate(context, event, self.PipelineDepth + 1)
        )
        return None

    async def generate(self, context, event, depth):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                        information of any data type with timestamp.

        depth : int


        """
        raise NotImplementedError()



import ctypes
import struct

libc6 = ctypes.cdll.LoadLibrary("libc.so.6")

# inotify constants
IN_ACCESS = 0x00000001  #: File was accessed
IN_MODIFY = 0x00000002  #: File was modified
IN_ATTRIB = 0x00000004  #: Metadata changed
IN_CLOSE_WRITE = 0x00000008  #: Writable file was closed
IN_CLOSE_NOWRITE = 0x00000010  #: Unwritable file closed
IN_OPEN = 0x00000020  #: File was opened
IN_MOVED_FROM = 0x00000040  #: File was moved from X
IN_MOVED_TO = 0x00000080  #: File was moved to Y
IN_CREATE = 0x00000100  #: Subfile was created
IN_DELETE = 0x00000200  #: Subfile was deleted
IN_DELETE_SELF = 0x00000400  #: Self was deleted
IN_MOVE_SELF = 0x00000800  #: Self was moved

IN_UNMOUNT = 0x00002000  #: Backing fs was unmounted
IN_Q_OVERFLOW = 0x00004000  #: Event queue overflowed
IN_IGNORED = 0x00008000  #: File was ignored

IN_ONLYDIR = 0x01000000  #: only watch the path if it is a directory
IN_DONT_FOLLOW = 0x02000000  #: don't follow a sym link
IN_EXCL_UNLINK = 0x04000000  #: exclude events on unlinked objects
IN_MASK_ADD = 0x20000000  #: add to the mask of an already existing watch
IN_ISDIR = 0x40000000  #: event occurred against dir
IN_ONESHOT = 0x80000000  #: only send event once

IN_ALL_EVENTS = (
    IN_MODIFY
    | IN_MOVED_FROM
    | IN_MOVED_TO
    | IN_CREATE
    | IN_DELETE
    | IN_DELETE_SELF
    | IN_MOVE_SELF
    | IN_Q_OVERFLOW
)

# inotify function prototypes
inotify_init = libc6.inotify_init
inotify_init.argtypes = []
inotify_init.restype = ctypes.c_int

inotify_add_watch = libc6.inotify_add_watch
inotify_add_watch.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_uint32]
inotify_add_watch.restype = ctypes.c_int


EVENT_FMT = "iIII"
EVENT_SIZE = struct.calcsize(EVENT_FMT)



from .analyzer import *
from .common import *
from .crypto import *
from .file import *
from .filter import *
from .kafka import *
from .matrix import *
from .declarative import *
from .integrity import *
from .test_config_defaults import *
from .test_metrics_service import *



import logging
import asyncio
from bspump.asab import Service


L = logging.getLogger(__file__)


class BSPumpService(Service):
    """
    Service registry based on Service object. Read more in ASAB documentation `Service <https://asab.readthedocs.io/en/latest/asab/service.html`_.

    """

    def __init__(self, app, service_name="bspump.PumpService"):
        """
        Initializes parameters passed to the Service class.

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_.

        service_name : str, Service name
                        string variable containing ""bspump.PumpService


        |

        """
        super().__init__(app, service_name)

        self.Pipelines = dict()
        self.Connections = dict()
        self.Lookups = dict()
        self.LookupFactories = []
        self.Matrixes = dict()

        self.App = app

    def locate(self, address):
        """
        Locates a pipeline, source or processor based on the addressed parameter.

        **Parameters**

        address : str, ID
                        Address of an pipeline component.
                        1. To locate a Pipeline, use the pipeline ID.
                        2. To locate a Processor or a Sink, use the pipeline ID and the component ID separated by a dot, like 'pipelineId.componentId'.
                        3. To locate a Source, add '*' before ID of the source, like 'pipeline.*SourceId'.

        **Returns**

        Pipeline, Source or Processor object if it is found in the components list, otherwise None.

        **Usage**

        >>> pipeline = service.locate("SuperCoolPipeline")
        >>> pipeline.Id
        SuperCoolPipeline

        >>> processor = service.locate("SuperCoolPipeline.PPrintProcessor")
        >>> processor.Id
        PPrintProcessor

        >>> source = service.locate("SuperCoolPipeline.*MySource")
        >>> source.Id
        MySource
        """
        if "." in address:
            p, t = address.split(".", 1)
        else:
            p = address
            t = None
        pipeline = self.Pipelines.get(p)
        if pipeline is None:
            return None
        elif t is None:
            return pipeline

        if t[:1] == "*":
            for source in pipeline.Sources:
                if source.Id == t[1:]:
                    return source
        else:
            for processor in pipeline.iter_processors():
                if processor.Id == t:
                    return processor

        return None

    # Pipelines

    def add_pipeline(self, pipeline):
        """
        Adds a pipeline to the BSPump.

        **Parameters**

        pipeline : Pipeline
                        Name of the Pipeline.

        """
        if pipeline.Id in self.Pipelines:
            raise RuntimeError(
                "Pipeline with id '{}' is already registered".format(pipeline.Id)
            )

        self.Pipelines[pipeline.Id] = pipeline

    def add_pipelines(self, *pipelines):
        """
        Adds a pipelines the BSPump.

        **Parameters**

        *pipelines : list
                        List of pipelines that are add to the BSPump.

        """
        for pipeline in pipelines:
            self.add_pipeline(pipeline)

    def del_pipeline(self, pipeline):
        """
        Deletes a pipeline from a list of Pipelines.

        **Parameters*

        pipeline : str, ID
                        ID of a pipeline.

        """
        del self.Pipelines[pipeline.Id]

    # Connections

    def add_connection(self, connection):
        """
        Adds a connection to the Connection dictionary.

        **Parameters**

        connection : str, ID
                        ID of a connection.


        :return: connection

        """
        if connection.Id in self.Connections:
            raise RuntimeError("Connection '{}' already created".format(connection.Id))
        self.Connections[connection.Id] = connection
        return connection

    def add_connections(self, *connections):
        """
        Adds a connections to the Connection dictionary.

        **Parameters**

        *connection : str, ID
                        list of IDs of a connections.

        """
        for connection in connections:
            self.add_connection(connection)

    def locate_connection(self, connection_id):
        """
        Locates connection based on connection ID.

        **Parameters**

        connection_id : ID
                        Connection ID.

        """
        from .abc.connection import Connection

        if isinstance(connection_id, Connection):
            return connection_id
        try:
            return self.Connections[connection_id]
        except KeyError:
            raise KeyError(
                "Cannot find connection id '{}' (did you call add_connection() before add_pipeline() ?)".format(
                    connection_id
                )
            )

    # Lookups

    def add_lookup(self, lookup):
        """
        Sets a lookup based on Lookup.

        **Parameters**

        lookup : Lookup
                        Name of the Lookup.

        :return: lookup

        """
        if lookup.Id in self.Lookups:
            raise RuntimeError("Lookup '{}' already created".format(lookup.Id))
        self.Lookups[lookup.Id] = lookup
        return lookup

    def add_lookups(self, *lookups):
        """
        Adds a list of lookups to the Pipeline.

        **Parameters**

        lookup : Lookup
                        List of Lookups.

        """
        for lookup in lookups:
            self.add_lookup(lookup)

    def locate_lookup(self, lookup_id, context=None):
        """
        Locates lookup based on ID.

        **Parameters**

        lookup_id : ID
                        ID of a Lookup.

        context : ,default = None
                        Additional information.

        :return: lookup from the lookup service or form the internal dictionary.

        """
        from .abc.lookup import Lookup

        if isinstance(lookup_id, Lookup):
            return lookup_id

        # TODO: Make sure the lookup is always properly returned
        # #1 - Return lookup from the lookup service
        for lookup_factory in self.LookupFactories:
            lookup = lookup_factory.locate_lookup(lookup_id, context)
            if lookup is not None:
                return lookup

        # #2 - Return lookup from the internal dictionary
        try:
            return self.Lookups[lookup_id]
        except KeyError:
            pass

        raise KeyError(
            "Cannot find lookup id '{}' (did you call add_lookup() ?)".format(lookup_id)
        )

    def add_lookup_factory(self, lookup_factory):
        """
        Adds a lookup factory

        **Parameters**

        lookup_factory :
                        Name of lookup factory.

        """
        self.LookupFactories.append(lookup_factory)

    # Matrixes

    def add_matrix(self, matrix):
        """
        Adds a matrix to the Pipeline.

        **Parameters**

        matrix : Matrix
                        Name of Matrix.

        :return: matrix

        """
        if matrix.Id in self.Matrixes:
            raise RuntimeError("Matrix '{}' already created".format(matrix.Id))

        self.Matrixes[matrix.Id] = matrix
        return matrix

    def add_matrixes(self, *matrixes):
        """
        Adds a list of Matrices to the Pipeline.

        **Parameters**

        *matrixes : list
                        List of matrices.

        """
        for matrix in matrixes:
            self.add_matrix(matrix)

    def locate_matrix(self, matrix_id):
        """
        Locates a matrix based on matrix ID

        **Parameters**

        matrix_id : str, ID
                        ID of a matrix.

        """
        from .matrix.matrix import Matrix

        if isinstance(matrix_id, Matrix):
            return matrix_id
        try:
            return self.Matrixes[matrix_id]
        except KeyError:
            raise KeyError(
                "Cannot find matrix id '{}' (did you call add_matrix() ?)".format(
                    matrix_id
                )
            )

    #

    async def initialize(self, app):
        """
        Initializes an Application based on ASAB `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_

        """
        # Run initialization of lookups
        lookup_update_tasks = []
        for lookup in self.Lookups.values():
            if not lookup.Lazy:
                lookup_update_tasks.append(lookup.ensure_future_update(app.Loop))

        # Await all lookups
        if len(lookup_update_tasks) > 0:
            done, pending = await asyncio.wait(lookup_update_tasks)

        # Start all pipelines
        for pipeline in self.Pipelines.values():
            pipeline.start()

        if self.App.MQTTService is not None:
            self.App.MQTTService.components_initialize()

    async def finalize(self, app):
        """
        Stops all the pipelines

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html>`_

        :return:
        """
        # Stop all started pipelines
        if len(self.Pipelines) > 0:
            await asyncio.gather(
                *[pipeline.stop() for pipeline in self.Pipelines.values()]
            )



from ..abc.lookup import MappingLookup
from ..abc.lookup import AsyncLookupMixin
from ..cache import CacheDict
import asyncio
import logging

###

L = logging.getLogger(__name__)

###


class MongoDBLookup(MappingLookup, AsyncLookupMixin):
    """
            The lookup that is linked with a MongoDB.
            It provides a mapping (dictionary-like) interface to pipelines.
            It feeds lookup data from Mongo using a query.
            It also has a simple cache to reduce a number of database hits.

            **configs**

            *database* - Mongo database name

            *collection* - Mongo collection name

            *key* - field name to match


            Example:

    .. code:: python

    The MongoDBLookup can be then located and used inside a custom enricher:

            class AsyncEnricher(bspump.Generator):

                    def __init__(self, app, pipeline, id=None, config=None):
                            super().__init__(app, pipeline, id, config)
                            svc = app.get_service("bspump.PumpService")
                            self.Lookup = svc.locate_lookup("MyMongoLookup")

                    async def generate(self, context, event, depth):
                            if 'user' not in event:
                                    return None

                            info = await self.Lookup.get(event['user'])

                            # Inject a new event into a next depth of the pipeline
                            self.Pipeline.inject(context, event, depth)

    """

    ConfigDefaults = {
        "database": "",  # Specify a database if you want to overload the connection setting
        "collection": "",  # Specify collection name
        "key": "",  # Specify key name used for search
        "changestream": False,
    }

    @classmethod
    def construct(cls, app, definition: dict, connection_id="MongoDBConnection"):
        """
        Usage:

                ---
                lookup: UserPasswordLookup
                config:
                        database: lookups
                        collection: userpassword
                        key: user
        """

        # TODO: Think about declarativity in lookups
        _id = definition.get("id", definition.get("declaration"))
        config = definition.get("config")
        svc = app.get_service("bspump.PumpService")
        mongodb_connection = svc.locate_connection(connection_id)
        return cls(app, connection=mongodb_connection, id=_id, config=config)

    def __init__(self, app, connection, id=None, config=None, cache=None):
        super().__init__(app, id=id, config=config)
        self.Connection = connection
        self.Database = self.Config["database"]
        self.Collection = self.Config["collection"]
        self.Key = self.Config["key"]
        self.Changestream = self.Config.getboolean("changestream")

        self.Loop = app.Loop
        self._changestream_future = None

        if len(self.Database) == 0:
            self.Database = self.Connection.Database

        if self.Changestream:
            self._changestream_future = asyncio.ensure_future(
                self._changestream(),
            )

        self.Count = -1
        if cache is None:
            self.Cache = CacheDict()
        else:
            self.Cache = cache

        metrics_service = app.get_service("asab.MetricsService")
        self.CacheCounter = metrics_service.create_counter(
            "mongodb.lookup", tags={}, init_values={"hit": 0, "miss": 0}
        )
        self.SuccessCounter = metrics_service.create_counter(
            "mongodb.lookup.success", tags={}, init_values={"hit": 0, "miss": 0}
        )

        app.PubSub.subscribe("Application.tick/10!", self._on_health_check)
        app.PubSub.subscribe("Application.exit!", self._on_exit)

    def _on_health_check(self, message_type):
        if self._changestream_future is not None and self.Changestream:
            # Future exists

            if not self._changestream_future.done():
                # Future didn't result yet
                # No sanitization needed
                return

            try:
                self._changestream_future.result()
            except Exception:
                # Connection future threw an error
                L.exception("Unexpected connection future error")

            # Future already resulted (with or without exception)
            self._changestream_future = None

        assert self._changestream_future is None

        if self.Changestream:
            self._changestream_future = asyncio.ensure_future(
                self._changestream(),
            )

    async def _on_exit(self, message_type):
        # On application exit, we cancel the future.
        if self._changestream_future is not None:
            self._changestream_future.cancel()

    def build_query(self, key):
        return {self.Key: key}

    async def _find_one(self, query):
        return await self.Connection.Client[self.Database][self.Collection].find_one(
            query
        )

    async def _changestream(self):
        try:
            async with self.Connection.Client[self.Database][
                self.Collection
            ].watch() as stream:
                async for change in stream:
                    self.Cache.clear()

        except asyncio.CancelledError:
            return

        except Exception as e:
            L.exception(f"Observed exception: {e}")
            await asyncio.sleep(5)

    async def get(self, key):
        """
        Obtain the value from lookup asynchronously.
        """
        try:
            value = self.Cache[key]
            self.CacheCounter.add("hit", 1)
        except KeyError:
            query = self.build_query(key)
            value = await self._find_one(query)
            if value is not None:
                self.Cache[key] = value
                self.CacheCounter.add("miss", 1)

        if value is None:
            self.SuccessCounter.add("miss", 1)
        else:
            self.SuccessCounter.add("hit", 1)
        return value

    async def _count(self, database):
        return await database[self.Collection].count_documents({})

    async def load(self):
        return True

    def __len__(self):
        return self.Count

    def __getitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        database = self.Connection.Client[self.Database].delegate
        self.Iterator = database[self.Collection].find()
        return self

    def __next__(self):
        element = self.Iterator.next()
        key = element.get(self.Key)
        if key is not None:
            self.Cache[key] = element
        return key



import asyncio
import datetime
import logging
import logging.handlers
import os
import pprint
import queue
import re
import socket
import sys
import time
import traceback
import urllib.parse

from .config import Config
from .timer import Timer
from .utils import running_in_container


LOG_NOTICE = 25
"""
Info log level that is visible in non-verbose mode. It should not be used for warnings and errors.
"""

logging.addLevelName(LOG_NOTICE, "NOTICE")

L = logging.getLogger(__name__)

_NAME_TO_LEVEL = {
    "NOTSET": logging.NOTSET,
    "NOT SET": logging.NOTSET,
    "NOT_SET": logging.NOTSET,
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "NOTICE": LOG_NOTICE,
    "LOG_NOTICE": LOG_NOTICE,
    "LOG NOTICE": LOG_NOTICE,
    "WARNING": logging.WARNING,
    "WARN": logging.WARNING,
    "ERROR": logging.ERROR,
    "FATAL": logging.CRITICAL,
    "CRITICAL": logging.CRITICAL,
}


class Logging(object):
    def __init__(self, app):
        self.RootLogger = logging.getLogger()

        self.ConsoleHandler = None
        self.FileHandler = None
        self.SyslogHandler = None

        if not self.RootLogger.hasHandlers():
            # Add console logger if needed
            if (
                os.isatty(sys.stdout.fileno())
                or os.environ.get("ASABFORCECONSOLE", "0") != "0"
            ):
                self._configure_console_logging()

            # Initialize file handler
            file_path = Config["logging:file"]["path"]

            if len(file_path) > 0:
                # Ensure file path
                directory = os.path.dirname(file_path)
                if not os.path.exists(directory):
                    os.makedirs(directory)

                self.FileHandler = logging.handlers.RotatingFileHandler(
                    file_path,
                    backupCount=Config.getint("logging:file", "backup_count"),
                    maxBytes=Config.getint("logging:file", "backup_max_bytes"),
                )
                self.FileHandler.setLevel(logging.DEBUG)
                self.FileHandler.setFormatter(
                    StructuredDataFormatter(
                        fmt=Config["logging:file"]["format"],
                        datefmt=Config["logging:file"]["datefmt"],
                        sd_id=Config["logging"]["sd_id"],
                    )
                )
                self.RootLogger.addHandler(self.FileHandler)

                rotate_every = Config.get("logging:file", "rotate_every")
                if rotate_every != "":
                    rotate_every = re.match(r"^([0-9]+)([dMHs])$", rotate_every)
                    if rotate_every is not None:
                        i, u = rotate_every.groups()
                        i = int(i)
                        if i <= 0:
                            self.RootLogger.error(
                                "Invalid 'rotate_every' configuration value."
                            )
                        else:
                            if u == "H":
                                i = i * 60 * 60
                            elif u == "M":
                                i = i * 60
                            elif u == "d":
                                i = i * 60 * 60 * 24
                            elif u == "s":
                                pass

                            # PubSub is not ready at this moment, we need to create timer in a future
                            async def schedule(app, interval):
                                self.LogRotatingTime = Timer(
                                    app, self._on_tick_rotate_check, autorestart=True
                                )
                                self.LogRotatingTime.start(i)

                            asyncio.ensure_future(schedule(app, i))

                    else:
                        self.RootLogger.error(
                            "Invalid 'rotate_every' configuration value."
                        )

            # Initialize syslog
            if Config["logging:syslog"].getboolean("enabled"):
                address = Config["logging:syslog"]["address"]

                if address[:1] == "/":
                    self.SyslogHandler = AsyncIOHandler(
                        app.Loop, socket.AF_UNIX, socket.SOCK_DGRAM, address
                    )

                else:
                    url = urllib.parse.urlparse(address)

                    if url.scheme == "tcp":
                        self.SyslogHandler = AsyncIOHandler(
                            app.Loop,
                            socket.AF_INET,
                            socket.SOCK_STREAM,
                            (
                                url.hostname
                                if url.hostname is not None
                                else "localhost",
                                url.port
                                if url.port is not None
                                else logging.handlers.SYSLOG_UDP_PORT,
                            ),
                        )

                    elif url.scheme == "udp":
                        self.SyslogHandler = AsyncIOHandler(
                            app.Loop,
                            socket.AF_INET,
                            socket.SOCK_DGRAM,
                            (
                                url.hostname
                                if url.hostname is not None
                                else "localhost",
                                url.port
                                if url.port is not None
                                else logging.handlers.SYSLOG_UDP_PORT,
                            ),
                        )

                    elif url.scheme == "unix-connect":
                        self.SyslogHandler = AsyncIOHandler(
                            app.Loop, socket.AF_UNIX, socket.SOCK_STREAM, url.path
                        )

                    elif url.scheme == "unix-sendto":
                        self.SyslogHandler = AsyncIOHandler(
                            app.Loop, socket.AF_UNIX, socket.SOCK_DGRAM, url.path
                        )

                    else:
                        self.RootLogger.warning(
                            "Invalid logging:syslog address '{}'".format(address)
                        )
                        address = None

                if self.SyslogHandler is not None:
                    self.SyslogHandler.setLevel(logging.DEBUG)
                    format = Config["logging:syslog"]["format"]
                    if format == "m":
                        self.SyslogHandler.setFormatter(
                            MacOSXSyslogFormatter(sd_id=Config["logging"]["sd_id"])
                        )
                    elif format == "5":
                        self.SyslogHandler.setFormatter(
                            SyslogRFC5424Formatter(sd_id=Config["logging"]["sd_id"])
                        )
                    elif format == "5micro":
                        self.SyslogHandler.setFormatter(
                            SyslogRFC5424microFormatter(
                                sd_id=Config["logging"]["sd_id"]
                            )
                        )
                    else:
                        self.SyslogHandler.setFormatter(
                            SyslogRFC3164Formatter(sd_id=Config["logging"]["sd_id"])
                        )
                    self.RootLogger.addHandler(self.SyslogHandler)

            # No logging is configured
            if (
                self.ConsoleHandler is None
                and self.FileHandler is None
                and self.SyslogHandler is None
            ):
                # Let's check if we run in Docker and if so, then log on stderr
                if running_in_container():
                    self._configure_console_logging()

        else:
            self.RootLogger.warning(
                "Logging seems to be already configured. Proceed with caution."
            )

        if Config["logging"].getboolean("verbose"):
            self.RootLogger.setLevel(logging.DEBUG)
        else:
            level_name = Config["logging"]["level"].upper()
            try:
                self.RootLogger.setLevel(_NAME_TO_LEVEL.get(level_name, level_name))
            except ValueError:
                L.error("Cannot detect logging level '{}'".format(level_name))

        # Fine-grained log level configurations
        levels = Config["logging"].get("levels")
        for level_line in levels.split("\n"):
            level_line = level_line.strip()
            if (
                len(level_line) == 0
                or level_line.startswith("#")
                or level_line.startswith(";")
            ):
                # line starts with a comment
                continue
            try:
                logger_name, level_name = level_line.split(" ", 1)
            except ValueError:
                L.error(
                    "Cannot read line '{}' in '[logging] levels' section, expected format: 'logger_name level_name'.".format(
                        level_line
                    )
                )
                continue
            level = _NAME_TO_LEVEL.get(level_name.upper(), level_name.upper())
            try:
                logging.getLogger(logger_name).setLevel(level)
            except ValueError:
                L.error(
                    "Cannot detect logging level '{}' for {} logger".format(
                        level_name, logger_name
                    )
                )

    def rotate(self):
        if self.FileHandler is not None:
            self.RootLogger.log(LOG_NOTICE, "Rotating logs")
            self.FileHandler.doRollover()

    async def _on_tick_rotate_check(self):
        if self.FileHandler is not None:
            if self.FileHandler.stream.tell() > 1000:
                self.rotate()

    def _configure_console_logging(self):
        self.ConsoleHandler = logging.StreamHandler(stream=sys.stderr)

        # Disable colors when running in container
        if running_in_container():
            self.ConsoleHandler.setFormatter(
                StructuredDataFormatter(
                    fmt=Config["logging:console"]["format"],
                    datefmt=Config["logging:console"]["datefmt"],
                    sd_id=Config["logging"]["sd_id"],
                    use_color=False,
                )
            )
        else:
            self.ConsoleHandler.setFormatter(
                StructuredDataFormatter(
                    fmt=Config["logging:console"]["format"],
                    datefmt=Config["logging:console"]["datefmt"],
                    sd_id=Config["logging"]["sd_id"],
                    use_color=True,
                )
            )

        self.ConsoleHandler.setLevel(logging.DEBUG)
        self.RootLogger.addHandler(self.ConsoleHandler)


class _StructuredDataLogger(logging.Logger):
    """
    This class extends a default python logger class, specifically by adding ``struct_data`` parameter to logging functions.
    It means that you can use expressions such as ``logger.info("Hello world!", struct_data={'key':'value'})``.
    """

    def _log(
        self,
        level,
        msg,
        args,
        exc_info=None,
        struct_data=None,
        extra=None,
        stack_info=False,
    ):
        if struct_data is not None:
            if extra is None:
                extra = dict()
            extra["_struct_data"] = struct_data

        super()._log(
            level, msg, args, exc_info=exc_info, extra=extra, stack_info=stack_info
        )


logging.setLoggerClass(_StructuredDataLogger)


class StructuredDataFormatter(logging.Formatter):
    """
    The logging formatter that renders log messages that includes structured data.
    """

    empty_sd = ""
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

    def __init__(
        self,
        facility=16,
        fmt=None,
        datefmt=None,
        style="%",
        sd_id="sd",
        use_color: bool = False,
    ):
        super().__init__(fmt, datefmt, style)
        self.SD_id = sd_id
        self.Facility = facility
        self.UseColor = use_color

    def format(self, record):
        """
        Format the specified record as text.
        """

        record.struct_data = self.render_struct_data(
            record.__dict__.get("_struct_data")
        )

        # The Priority value is calculated by first multiplying the Facility number by 8 and then adding the numerical value of the Severity.
        if record.levelno <= logging.DEBUG:
            severity = 7  # Debug
            color = self.BLUE
        elif record.levelno <= logging.INFO:
            severity = 6  # Informational
            color = self.GREEN
        elif record.levelno <= LOG_NOTICE:
            severity = 5  # Notice
            color = self.CYAN
        elif record.levelno <= logging.WARNING:
            severity = 4  # Warning
            color = self.YELLOW
        elif record.levelno <= logging.ERROR:
            severity = 3  # Error
            color = self.RED
        elif record.levelno <= logging.CRITICAL:
            severity = 2  # Critical
            color = self.MAGENTA
        else:
            severity = 1  # Alert
            color = self.WHITE

        if self.UseColor:
            levelname = record.levelname
            levelname_color = _COLOR_SEQ % (30 + color) + levelname + _RESET_SEQ
            record.levelname = levelname_color

        record.priority = (self.Facility << 3) + severity
        return super().format(record)

    def formatTime(self, record, datefmt=None):
        """
        Return the creation time of the specified LogRecord as formatted text.
        """

        ct = datetime.datetime.fromtimestamp(record.created)
        if datefmt is not None:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s.%03d" % (t, record.msecs)
        return s

    def render_struct_data(self, struct_data):
        """
        Return the string with structured data.
        """

        if struct_data is None:
            return self.empty_sd

        else:
            return "[{sd_id} {sd_params}] ".format(
                sd_id=self.SD_id,
                sd_params=" ".join(
                    ['{}="{}"'.format(key, val) for key, val in struct_data.items()]
                ),
            )


def _loop_exception_handler(loop, context):
    """
    This is an logging exception handler for asyncio.
    It's purpose is to nicely log any unhandled excpetion that arises in the asyncio tasks.
    """

    exception = context.pop("exception", None)

    message = context.pop("message", "")
    if len(message) > 0:
        message += "\n"

    if len(context) > 0:
        message += pprint.pformat(context)

    if exception is not None:
        ex_traceback = exception.__traceback__
        tb_lines = [
            line.rstrip("\n")
            for line in traceback.format_exception(
                exception.__class__, exception, ex_traceback
            )
        ]
        message += "\n" + "\n".join(tb_lines)

    logging.getLogger().error(message)


class MacOSXSyslogFormatter(StructuredDataFormatter):
    """
    It implements Syslog formatting for Mac OSX syslog (aka format ``m``).
    """

    def __init__(self, fmt=None, datefmt=None, style="%", sd_id="sd"):
        fmt = "<%(priority)s>%(asctime)s {app_name}[{proc_id}]: %(levelname)s %(name)s %(struct_data)s%(message)s\000".format(
            app_name=Config["logging"]["app_name"],
            proc_id=os.getpid(),
        )

        # Initialize formatter
        super().__init__(fmt=fmt, datefmt="%b %d %H:%M:%S", style=style, sd_id=sd_id)


class SyslogRFC3164Formatter(StructuredDataFormatter):
    """
    Implementation of a legacy or BSD Syslog (RFC 3164) formatting (aka format ``3``).
    """

    def __init__(self, fmt=None, datefmt=None, style="%", sd_id="sd"):
        fmt = "<%(priority)s>%(asctime)s {hostname} {app_name}[{proc_id}]:%(levelname)s %(name)s %(struct_data)s%(message)s\000".format(
            app_name=Config["logging"]["app_name"],
            hostname=socket.gethostname(),
            proc_id=os.getpid(),
        )

        # Initialize formatter
        super().__init__(fmt=fmt, datefmt="%b %d %H:%M:%S", style=style, sd_id=sd_id)


class SyslogRFC5424Formatter(StructuredDataFormatter):
    """
    It implements Syslog formatting for Mac OSX syslog (aka format ``5``).
    """

    empty_sd = " "

    def __init__(self, fmt=None, datefmt=None, style="%", sd_id="sd"):
        fmt = '<%(priority)s>1 %(asctime)s.%(msecs)dZ {hostname} {app_name} {proc_id} %(name)s [log l="%(levelname)s"]%(struct_data)s%(message)s'.format(
            app_name=Config["logging"]["app_name"],
            hostname=socket.gethostname(),
            proc_id=os.getpid(),
        )

        # Initialize formatter
        super().__init__(fmt=fmt, datefmt="%Y-%m-%dT%H:%M:%S", style=style, sd_id=sd_id)

        # Convert time to GMT
        self.converter = time.gmtime


class SyslogRFC5424microFormatter(StructuredDataFormatter):
    """
    It implements Syslog formatting for syslog (aka format ``micro``) in RFC5424micro format.
    """

    empty_sd = "-"

    def __init__(self, fmt=None, datefmt=None, style="%", sd_id="sd"):
        fmt = '<%(priority)s>1 %(asctime)sZ {hostname} {app_name} {proc_id} %(name)s [log l="%(levelname)s"]%(struct_data)s%(message)s'.format(
            app_name=Config["logging"]["app_name"],
            hostname=socket.gethostname(),
            proc_id=os.getpid(),
        )

        super().__init__(
            fmt=fmt, datefmt="%Y-%m-%dT%H:%M:%S.%f", style=style, sd_id=sd_id
        )
        self.converter = time.gmtime


class AsyncIOHandler(logging.Handler):
    """
    A logging handler similar to a standard `logging.handlers.SocketHandler` that utilizes `asyncio`.
    It implements a queue for decoupling logging from a networking. The networking is fully event-driven via `asyncio` mechanisms.
    """

    def __init__(
        self,
        loop,
        family,
        sock_type,
        address,
        facility=logging.handlers.SysLogHandler.LOG_LOCAL1,
    ):
        logging.Handler.__init__(self)

        self._family = family
        self._type = sock_type
        self._address = address
        self._loop = loop

        self._socket = None
        self._reset()

        self._queue = queue.Queue()

        self._loop.call_soon(self._connect, self._loop)

    def _reset(self):
        self._write_ready = False
        if self._socket is not None:
            self._loop.remove_writer(self._socket)
            self._loop.remove_reader(self._socket)
            self._socket.close()
            self._socket = None

    def _connect(self, loop):
        self._reset()

        self._socket = socket.socket(self._family, self._type)
        self._socket.setblocking(0)
        self._socket.connect(self._address)

        self._loop.add_writer(self._socket, self._on_write)
        self._loop.add_reader(self._socket, self._on_read)

    def _on_write(self):
        self._write_ready = True
        self._loop.remove_writer(self._socket)

        while not self._queue.empty():
            msg = self._queue.get_nowait()
            self._socket.sendall(msg)

    def _on_read(self):
        try:
            _ = self._socket.recvfrom(1024)
            # We receive "something" ... let's ignore that!
            return
        except Exception as e:
            print(
                "Error on the syslog socket '{}'".format(self._address),
                e,
                file=sys.stderr,
            )

        # Close a socket - there is no reason for reading or socket is actually closed
        self._reset()

    def emit(self, record):
        """
        This is the entry point for log entries.
        """
        try:
            msg = self.format(record).encode("utf-8")

            if self._write_ready:
                try:
                    self._socket.sendall(msg)
                except Exception as e:
                    print(
                        "Error when writing to syslog '{}'".format(self._address),
                        e,
                        file=sys.stderr,
                    )
                    self._enqueue(msg)

            else:
                self._enqueue(msg)

        except Exception as e:
            print(
                "Error when emit to syslog '{}'".format(self._address),
                e,
                file=sys.stderr,
            )
            self.handleError(record)

    def _enqueue(self, record):
        self._queue.put(record)


_RESET_SEQ = "\033[0m"
_COLOR_SEQ = "\033[1;%dm"
_BOLD_SEQ = "\033[1m"



from .connection import MQTTConnection
from .source import MQTTSource
from .sink import MQTTSink
from .service import MQTTService

__all__ = ("MQTTConnection", "MQTTSource", "MQTTSink", "MQTTService")



import logging

from .abc import Expression
from .expression.value.valueexpr import VALUE

###

L = logging.getLogger(__name__)

###


class ExpressionOptimizer(object):
    """
    Optimizes an expression using walk strategy and individual optimize methods.
    """

    def __init__(self, app):
        self.App = app

    def optimize(self, expression):
        # We run optimizations till we finish tree walk without any optimization found
        retry = True
        counter = 0
        while retry:
            retry = False

            counter += 1

            if not isinstance(expression, Expression):
                expression = VALUE(self.App, value=expression)

            # Walk the syntax tree
            for parent, key, obj in expression.walk():
                if not isinstance(obj, Expression):
                    continue

                # Check if the node could be optimized
                opt_obj = obj.optimize()
                if opt_obj is None:
                    continue

                assert obj is not opt_obj

                if parent is None:
                    expression = opt_obj
                else:
                    # If yes, replace a given node by the optimized variant
                    parent.set(key, opt_obj)

                if counter > 100000:
                    raise RuntimeError(
                        "Optimization likely stucked at '{}'-'{}'-'{}'/'{}'".format(
                            parent, key, obj, opt_obj
                        )
                    )

                # ... and start again
                retry = True
                break

        return expression

    def optimize_many(self, expressions):
        return [self.optimize(expression) for expression in expressions]



import logging
import asyncio

from ..abc.sink import Sink

#

L = logging.getLogger(__name__)

#


class AsyncSink(Sink):
    """
    Inherit from the sink and override the `consume` coroutine:

    class MySink(bspump.aio.AsyncSink):

            async def consume(self, context, event):
                    ...
    """

    ConfigDefaults = {
        "output_queue_max_size": 100,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)

        self.Queue = asyncio.Queue()
        self.QueueMaxSize = int(self.Config["output_queue_max_size"])
        assert self.QueueMaxSize >= 1

        self.Task = asyncio.ensure_future(self._task())

        self.Exiting = False

        app.PubSub.subscribe("Application.exit!", self._on_exit)

    async def _task(self):
        while not self.Exiting:
            context, event = await self.Queue.get()

            # Unthrottle if needed
            if self.Queue.qsize() == (self.QueueMaxSize - 1):
                self.Pipeline.throttle(self.Queue, False)

            # Consume the event
            await self.consume(context, event)

    async def _on_exit(self, event_name):
        self.Exiting = True

        if self.Task is not None:
            t = self.Task
            self.Task = None
            t.cancel()
            await t

    def process(self, context, event):
        self.Queue.put_nowait((context, event))

        if self.Queue.qsize() == self.QueueMaxSize:
            self.Pipeline.throttle(self.Queue, True)

    async def consume(self, context, event):
        """
        Method to be overridden.
        """
        pass



import bspump.analyzer
import bspump.unittest


class TestGeoAnalyzer(bspump.unittest.ProcessorTestCase):
    def test_geo_analyzer(self):
        events = [
            (None, {"lat": 70, "lon": 10}),
            (None, {"lat": 50, "lon": 30}),
        ]
        self.set_up_processor(bspump.analyzer.GeoAnalyzer)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output],
            [{"lat": 70, "lon": 10}, {"lat": 50, "lon": 30}],
        )

        # TODO test self.Pipeline.Processor.Matrix



import json
import os
import ast
import sys
import tempfile
import re

from bspump.jupyter import *  # noqa: F403
import bspump.jupyter

config = None
__bitswan_dev = False
__bs_step_locals = {}


def contains_function_call(ast_tree, function_name):
    for node in ast.walk(ast_tree):
        if isinstance(node, ast.Call):  # Check if the node is a function call
            if isinstance(node.func, ast.Name) and node.func.id == function_name:
                return True
    return False


def indent_code(lines: list[str]) -> list[str]:
    multiline = False
    double_quotes = False
    indent_lines = []
    lines_out = []
    for i, line in enumerate(lines):
        if not multiline and line.strip(" ") != "":
            indent_lines.append(i)
        if '"""' in line or "'''" in line:
            if not multiline:
                double_quotes = '"""' in line
            elif (double_quotes and "'''" in line) or (
                not double_quotes and '"""' in line
            ):
                continue
            multiline = not multiline
            continue
    for i in range(len(lines)):
        _indent = "    " if i in indent_lines else ""
        lines_out.append(_indent + lines[i])
    return lines_out


class NotebookCompiler:
    _in_autopipeline = False
    _cell_number: int = 0
    _cell_processor_contents: dict[int, str] = {}

    def parse_cell(self, cell, fout):
        if cell["cell_type"] == "code":
            source = cell["source"]
            if len(source) > 0 and "#ignore" not in source[0]:
                code = (
                    "".join(cell["source"])
                    if isinstance(cell["source"], list)
                    else cell["source"]
                )

                clean_code = (
                    "\n".join(
                        [
                            re.sub(r"^\t+(?=\S)", "", line)
                            if not line.startswith("!")
                            else ""
                            for line in code.split("\n")
                        ]
                    ).strip("\n")
                    + "\n"
                )

                if not clean_code.strip():
                    return
                if not self._in_autopipeline:
                    fout.write(clean_code + "\n\n")
                else:
                    self._cell_processor_contents[self._cell_number] = (
                        "\n".join(indent_code(clean_code.split("\n"))) + "\n\n"
                    )
                if not self._in_autopipeline and contains_function_call(
                    ast.parse(clean_code), "auto_pipeline"
                ):
                    self._in_autopipeline = True

    def compile_notebook(self, ntb, out_path="tmp.py"):
        self._cell_number = 0
        self._in_autopipeline = False
        self._cell_processor_contents = {}
        with open(out_path, "w") as f:
            for cell in ntb["cells"]:
                self._cell_number += 1
                self.parse_cell(cell, f)
            step_func_code = f"""@async_step
async def processor_internal(inject, event):
{''.join(list(self._cell_processor_contents.values()))}    await inject(event)
"""
            f.write(step_func_code)


def main():
    app = App()  # noqa: F405
    compiler = NotebookCompiler()

    if app.Test:
        bspump.jupyter.bitswan_test_mode.append(True)

    with tempfile.TemporaryDirectory() as tmpdirname:
        if os.path.exists(app.Notebook):
            with open(app.Notebook) as nb:
                notebook = json.load(nb)
                compiler.compile_notebook(
                    notebook, out_path=f"{tmpdirname}/autopipeline_tmp.py"
                )
                sys.path.insert(0, tmpdirname)
                tmp_module = __import__("autopipeline_tmp")  # noqa: F841
        else:
            exit(f"Notebook {app.Notebook} not found")

        if bspump.jupyter.bitswan_auto_pipeline.get("sink") is not None:
            register_sink(  # noqa: F405
                bspump.jupyter.bitswan_auto_pipeline.get("sink")
            )  # noqa: F405
            end_pipeline()  # noqa: F405

        app.init_componets()
        app.run()


if __name__ == "__main__":
    main()



import orjson
import logging


L = logging.getLogger(__name__)

"""
Data feeders to be used in ElasticSearchSink.
"""


def data_feeder_create_or_index(event, _id):
    """
    Creates an index.

    **Parameters**

    event : Data with time stamp stored in any data type usually is in JSON.
                    You can specify an event that is passed to the method.

    """
    if _id is None:
        yield b'{"create":{}}\n'
    else:
        yield orjson.dumps({"index": {"_id": _id}}, option=orjson.OPT_APPEND_NEWLINE)

    yield orjson.dumps(event, option=orjson.OPT_APPEND_NEWLINE)


def data_feeder_create(event, _id):
    """
    Creates a data feeder.

    **Parameters**

    event : Data with time stamp stored in any data type usually is in JSON.
                    You can specify an event that is passed to the method.

    """
    if _id is None:
        yield b'{"create":{}}\n'
    else:
        yield orjson.dumps({"create": {"_id": _id}}, option=orjson.OPT_APPEND_NEWLINE)

    yield orjson.dumps(event, option=orjson.OPT_APPEND_NEWLINE)


def data_feeder_index(event, _id):
    """
    Description:

    **Parameters**

    event : Data with time stamp stored in any data type usually is in JSON.
                    You can specify an event that is passed to the method.

    """

    if _id is None:
        yield b'{"index":{}}\n'
    else:
        yield orjson.dumps({"index": {"_id": _id}}, option=orjson.OPT_APPEND_NEWLINE)

    yield orjson.dumps(event, option=orjson.OPT_APPEND_NEWLINE)


def data_feeder_update(event, _id):
    """
    Updates data feeder.

    **Parameters**

    event : Data with time stamp stored in any data type usually is in JSON.
                    You can specify an event that is passed to the method.

    """

    assert (
        _id is not None
    ), "_id must be present in the event when updating a document in ElasticSearch"

    yield orjson.dumps({"update": {"_id": _id}}, option=orjson.OPT_APPEND_NEWLINE)

    yield orjson.dumps({"doc": event}, option=orjson.OPT_APPEND_NEWLINE)


def data_feeder_delete(event, _id):
    """
    Deletes data feeder.

    **Parameters**

    event : Data with time stamp stored in any data type usually is in JSON.
                    You can specify an event that is passed to the method.

    """
    assert (
        _id is not None
    ), "_id must be present in the event when deleting a document from ElasticSearch"

    yield orjson.dumps({"delete": {"_id": _id}}, option=orjson.OPT_APPEND_NEWLINE)

    assert (
        len(event) == 0
    ), "When deleting items from ElasticSearch, no data should be provide, but '{}' found.".format(
        event
    )



import abc
import random


class Trigger(abc.ABC):
    """
    Abstract base class for all trigger types in BitSwan.

    Triggers are components that fire events at specific times or under
    specific conditions. When a trigger fires, it notifies all attached
    sources to execute their cycle() method.

    Key features:
    - Manages a set of attached sources
    - Handles firing triggers (notifying sources)
    - Supports pause/resume functionality
    - Tracks trigger state and timing
    - Supports maximum triggered source limits

    Usage:
        trigger = SomeTrigger(app)
        source = MyTriggerSource(app, pipeline)
        source.on(trigger)  # Attach source to trigger
    """

    def __init__(self, app, max_triggered=None, id=None):
        self.Id = id if id is not None else self.__class__.__name__
        self.Sources = set()
        self.Paused = False
        self.LastFireAt = 0
        self.Loop = app.Loop
        self._max_triggered = max_triggered

    def add(self, source):
        self.Sources.add(source)

    def remove(self, source):
        self.Sources.remove(source)

    def fire(self):
        if self.Paused:
            return

        self.LastFireAt = self.Loop.time()

        if self._max_triggered is None:
            for source in self.Sources:
                source.TriggerEvent.set()

        else:
            # Maximum number of triggered event is defined
            triggered = []
            untriggered = []
            for source in self.Sources:
                if source.TriggerEvent.is_set():
                    triggered.append(source)
                else:
                    untriggered.append(source)
            to_trigger = self._max_triggered - len(triggered)
            random.shuffle(untriggered)

            while to_trigger > 0:
                if len(untriggered) == 0:
                    break
                source = untriggered.pop()
                source.TriggerEvent.set()
                to_trigger -= 1

    def done(self, trigger_source):
        """
        Called by TriggerSource when cycle is completed.
        """
        pass

    def pause(self, pause=True):
        self.Paused = pause
        for source in self.Sources:
            source.Pipeline.PubSub.publish(
                "bspump.trigger.pause!", pause=pause, asynchronously=True
            )



import logging
import functools
import itertools

from .declerror import DeclarationError

#

L = logging.getLogger(__name__)

#

_IdGenerator = itertools.count(1)


class Expression(object):
    Attributes = (
        None  # Must be a list, if missing, then the Expression class is not complete
    )

    Category = "Others"

    def __init__(self, app):
        if self.Attributes is None:
            raise NotImplementedError("Missing 'Attributes' in {}".format(self))

        self.Id = "{}.E{}".format(self.__class__.__name__, next(_IdGenerator))
        self.App = app
        self.Location = None
        self.Node = None  # The YAML node, assigned by a builder during YAML parsing

    def __call__(self, context, event, *args, **kwargs):
        raise RuntimeError("Call of the abstract method")

    def set_location(self, location):
        self.Location = location

    def get_location(self):
        return self.Location

    def initialize(self):
        """
        Initialize the expression after the syntax tree was build.
        :return:
        """
        return

    def optimize(self):
        return None

    def walk(self, parent=None, key=None):
        """
        key in the Expression is the string, which is a name of the attribute with the child expression
        """

        yield (parent, key, self)

        # Attributes may change during iteration, which could produce an error
        current_attributes = self.Attributes.copy()

        for key in current_attributes:
            v = getattr(self, key, None)

            if isinstance(v, Expression):
                for x in v.walk(self, key):
                    yield x
            else:
                yield (self, key, v)

    def set(self, key, value):
        setattr(self, key, value)

    def get_outlet_type(self):
        return "???"

    def consult_inlet_type(self, key, child):
        raise NotImplementedError("Parent consultation at '{}'".format(self))


class SequenceExpression(Expression):
    """
    Convenient class that fits onto SequenceNode from Yaml.

    Example of __call__() method implementation.

    def __call__(self, context, event, *args, **kwargs):
            return functools.reduce(
                    lambda x, y: evaluate(x, context, event, *args, **kwargs) + evaluate(y, context, event, *args, **kwargs),
                    self.Items
            )
    """

    Attributes = {"Items": []}

    def __init__(self, app, *, sequence):
        super().__init__(app)
        self.Items = []

        for i in sequence:
            if isinstance(i, Expression):
                self.Items.append(i)

            else:
                assert (
                    isinstance(i, (int, str, float, bytes, bool, tuple, list, dict))
                    or i is None
                )

                # So far used in the WHEN expression
                if isinstance(i, dict):
                    continue

                from .expression import VALUE

                self.Items.append(VALUE(app, value=i))

    def reduce(self, operator, context, event, *args, **kwargs):
        iterator = [
            evaluate(item, context, event, *args, **kwargs) for item in self.Items
        ]
        return functools.reduce(operator, iterator)

    def walk(self, parent=None, key=None):
        """
        key in the Sequence expression is the integer, which is a position in self.Items list
        """
        yield (parent, key, self)

        for key, i in enumerate(self.Items):
            if isinstance(i, Expression):
                for x in i.walk(self, key):
                    yield x
            else:
                raise NotImplementedError(":-(")

    def set(self, key, value):
        self.Items[key] = value

    def get_items_inlet_type(self):
        """
        This method evaluate the inlet type for `Items` sequence
        """
        return "?"


def evaluate(value, context, event, *args, **kwargs):
    """
    This is OBSOLETED method and will be removed.
    DO NOT USE IT!
    """
    try:
        return value(context, event, *args, **kwargs)
    except Exception as e:
        raise DeclarationError(original_exception=e, location=value.get_location())



import collections
import logging

import mongoquery

from .analyzer import Analyzer

###

L = logging.getLogger(__name__)


###


class LatchAnalyzer(Analyzer):
    """
    The `LatchAnalyzer` accumulates events in the `Latch` variable.
    The `Latch` is a queue of maximum size specified in configuration - `latch_max_size`

    If `latch_max_size` is 0 then `Latch` is not limited

    If accumulated events exceeds `latch_max_size` then first event is dropped.

    `Latch` can be filled based on the `query` variable (True by default).
    The query may be:
    1. `True`, then all events will be added to `Latch`.
    2. `False`, all events will be skipped.
    3. Dictionary, following the mongo-like query syntaxis (see the rules in `ContentFilter`).
    In this case only events matched with this query will be added to the `Latch`.

    The query can be injected with an API call to allow to control events in the latch.

    """

    ConfigDefaults = {
        "latch_max_size": 50,  # 0 means unlimited size
    }

    def __init__(
        self,
        app,
        pipeline,
        query=True,
        analyze_on_clock=False,
        inclusive=False,
        id=None,
        config=None,
    ):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        pipeline : Pipeline
                Name of the Pipeline.

        query : bool, default = True
                description

        analyze_on_clock : bool, default = False

        inclusive : bool, default = False

        id : str, default = None

        config : JSON, default = None
                configuration file with additional information.

        """
        super().__init__(
            app, pipeline, analyze_on_clock=analyze_on_clock, id=id, config=config
        )
        max_size = int(self.Config.get("latch_max_size"))
        if max_size == 0:
            self.Latch = collections.deque()
        else:
            self.Latch = collections.deque(maxlen=max_size)

        # Check if the query is correctly implemented
        if isinstance(query, bool):
            self.Query = query
        else:
            try:
                self.Query = mongoquery.Query(query)
                self.Query.match({})
            except mongoquery.QueryError:
                L.warning("Incorrect query")
                raise

        self.Inclusive = inclusive

    def process(self, context, event):
        """
        Description:

        **Parameters**

        context :

        event : any data type
                information with timestamp.

        :return: event
        """
        if self.Query is True:
            self.Latch.append(event)

        elif self.Query is False:
            return event

        elif self.Query.match(event) != self.Inclusive:
            self.Latch.append(event)
        return event



import fnmatch
import glob
import logging
import os.path
import platform

#

L = logging.getLogger(__file__)

#


if platform.system() == "Windows":

    def _is_file_open(fname):
        # TODO: Provide implementation of _is_file_open() for Windows
        return False

else:

    def _is_file_open(fname):
        # TODO: Provide implementation of _is_file_open() for Linux
        # result = subprocess.run(['lsof', fname], stdout=subprocess.PIPE)
        # return len(result.stdout) != 0
        return False


def iter_files_glob(path, gauge, loop, exclude="", include=""):
    if path is None:
        return []
    if path == "":
        return []
    filelist = glob.glob(path, recursive=True)
    filelist.sort()

    # We clone the file list and send it to file_check which logs statistics about the files
    filelist_to_check = []
    filelist_to_check.extend(filelist)
    loop.call_soon_threadsafe(_file_check, filelist_to_check, gauge)

    for fname in filelist:
        if any(
            [
                fname.endswith("-locked"),
                fname.endswith("-failed"),
                fname.endswith("-processed"),
                not os.path.isfile(fname),
            ]
        ):
            continue

        if exclude != "":
            if fnmatch.fnmatch(fname, exclude):
                continue
        if include != "":
            if not fnmatch.fnmatch(fname, include):
                continue

        if _is_file_open(fname):
            continue
        yield fname
    return


def _glob_scan(path, gauge, loop, exclude="", include=""):
    try:
        return iter_files_glob(path, gauge, loop, exclude, include).__next__()
    except StopIteration:
        return None


def _file_check(filelist, gauge):
    file_count = {
        "processed": 0,
        "unprocessed": 0,
        "failed": 0,
        "locked": 0,
        "all_files": 0,
    }

    file_count["all_files"] += len(filelist)
    for file in filelist:
        if file.endswith("-locked"):
            file_count["locked"] += 1
            continue
        if file.endswith("-failed"):
            file_count["failed"] += 1
            continue
        if file.endswith("-processed"):
            file_count["processed"] += 1
            continue

        file_count["unprocessed"] += 1

    gauge.set("processed", file_count["processed"])
    gauge.set("failed", file_count["failed"])
    gauge.set("locked", file_count["locked"])
    gauge.set("unprocessed", file_count["unprocessed"])
    gauge.set("all_files", file_count["all_files"])


def test_iter_files_glob():
    class FakeGauge:
        def set(self, key, value):
            pass

    fake_gauge = FakeGauge()

    class FakeLoop:
        def call_soon_threadsafe(self, func, *args):
            func(*args)

    fake_loop = FakeLoop()
    filelist = list(
        iter_files_glob("bspump/test-data/globscan/*", fake_gauge, fake_loop)
    )
    assert len(filelist) == 3
    assert filelist[0] == "bspump/test-data/globscan/1"
    assert filelist[1] == "bspump/test-data/globscan/2"
    assert filelist[2] == "bspump/test-data/globscan/3"

    # test _glob_scan
    assert (
        _glob_scan("bspump/test-data/globscan/*", fake_gauge, fake_loop)
        == "bspump/test-data/globscan/1"
    )



import bspump.analyzer
import bspump.unittest
import mongoquery


class CustomSmallLatchAnalyzer(bspump.analyzer.LatchAnalyzer):
    ConfigDefaults = {
        "latch_max_size": 2,
    }


class CustomInfiniteLatchAnalyzer(bspump.analyzer.LatchAnalyzer):
    ConfigDefaults = {
        "latch_max_size": 0,
    }


class TestLatch(bspump.unittest.ProcessorTestCase):
    def test_analyzer(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(CustomSmallLatchAnalyzer)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )
        self.assertEqual(len(self.Pipeline.Processor.Latch), 2)

    def test_analyzer_query_false(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(CustomSmallLatchAnalyzer, query=False)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )
        self.assertEqual(len(self.Pipeline.Processor.Latch), 0)

    def test_infinite_analyzer_query_false(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(CustomInfiniteLatchAnalyzer, query=False)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )
        self.assertEqual(len(self.Pipeline.Processor.Latch), 0)
        self.assertIsNone(self.Pipeline.Processor.Latch.maxlen)

    def test_analyzer_with_mongo_query(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(CustomSmallLatchAnalyzer, query={"foo": "bar"})

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )
        self.assertEqual(list(self.Pipeline.Processor.Latch), [{"foo": "bar"}])

    def test_analyzer_with_mongo_query_inclusive(self):
        events = [
            (None, {"foo": "bar"}),
            (None, {"fizz": "buzz"}),
        ]
        self.set_up_processor(
            CustomSmallLatchAnalyzer, query={"foo": "bar"}, inclusive=True
        )

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"foo": "bar"}, {"fizz": "buzz"}]
        )
        self.assertEqual(list(self.Pipeline.Processor.Latch), [{"fizz": "buzz"}])

    def test_analyzer_with_invalid_query(self):
        with self.assertRaises(mongoquery.QueryError):
            self.set_up_processor(bspump.analyzer.LatchAnalyzer, query={"$foo": 2})



import time
import hashlib
import collections


class Session(collections.MutableMapping):

    """Session dict-like object."""

    def __init__(self, storage, id, new, max_age=None):
        self._id = id
        self._changed = False
        self._mapping = {}
        self._new = new
        self.MaxAge = max_age
        self.Storage = storage

        self._created = int(time.time())

    def __hash__(self):
        return hash((self.__class__, self._id))

    def __repr__(self):
        return "<{} [{}, new:{}, changed:{}, created:{} expired:{}] {!r}>".format(
            self.__class__.__name__,
            hashlib.sha224(self._id.encode("utf-8")).hexdigest()
            if self._id is not None
            else "-",
            self._new,
            self._changed,
            self._created,
            self.is_expired(),
            self._mapping,
        )

    def is_expired(self):
        expiration = self._created + self.MaxAge
        return time.time() >= expiration

    def reset(self):
        self._new = False
        self._change = False

    @property
    def Id(self):
        return self._id

    def set_id(self, id):
        if self._id is not None or not self._new:
            raise RuntimeError("Session id is already set!")
        self._id = id

    @property
    def Created(self):
        return self._created

    def is_new(self):
        return self._new

    def mark_changed(self):
        self._changed = True

    def __len__(self):
        return len(self._mapping)

    def __iter__(self):
        return iter(self._mapping)

    def __contains__(self, key):
        return key in self._mapping

    def __getitem__(self, key):
        return self._mapping[key]

    def __setitem__(self, key, value):
        self._mapping[key] = value
        self._changed = True

    def __delitem__(self, key):
        del self._mapping[key]
        self._changed = True



from .opportunistic import OpportunisticTrigger
from .periodic import PeriodicTrigger
from .pubsub import PubSubTrigger
from .runonce import RunOnceTrigger
from .trigger import Trigger
from .crontrig import CronTrigger

__all__ = (
    "Trigger",
    "OpportunisticTrigger",
    "RunOnceTrigger",
    "PubSubTrigger",
    "PeriodicTrigger",
    "CronTrigger",
)
